//! this example demonstrates the type of four-core clocked inverter conceived
//! of in the https://uninsane.org/ blogpost.
//!
//! we're going for something like this:
//!
//! ```
//!          ____         ____         ____         ____
//!  + ___  /    \ +___  /    \ +___  /    \ +___  /    \  ___ +
//! IN ___ |o    o| ___ |x    o| ___ |x    o| ___ |x    o| ___ OUT
//!  -      \_o__/ -     \_o__/ -     \_o__/ -     \_o__/      -
//!           ||           ||           ||           ||
//!          +  -         +  -         +  -         +  -
//!          CTL0         CTL1         CTL2         CTL3
//! ```
//!
//! conventions:
//! - CW core polarization is logic '1'.
//! - differential signals are assumed to flow clockwise. i.e. left-to-right, with + on top and -
//!   on bottom.
//! - 'x' denotes a signal going "into" the page. 'o' for a signal going "out of" the page.
//! - the core driven by CTLn is named Sn, interchangeably used to represent the state (0 or 1) of
//!   that core
//!
//! hence:
//! - the input into core S0 has current flowing out of the page and tends to drive S0 to logic '0'.
//! - the input into S1, S2, S3 tends these cores to logic '1'.
//! - a positive current to CTLn will "clear" Sn to the logic '0' state.
//!
//! in this device, S0 is effectively an inverter, with S1, S2, S3 acting as buffers.
//! for the purpose of this simulation, all terminal wires are closed loops either explicitly
//! driven or measured.
//!
//! to run this, from toplevel directory:
//! ```
//! $ cargo run --release --bin multi_core_inverter
//! $ pushd crates/coremem; cargo run --release --bin viewer ../../out/applications/multi_core_inverter/0/ ; popd
//! ```

use coremem::geom::{Meters, Torus};
use coremem::sim::units::Seconds;
use coremem::mat::{Ferroxcube3R1MH, IsoConductorOr, IsomorphicConductor};
use coremem::{Driver, SpirvDriver};

type Mat = IsoConductorOr<f32, Ferroxcube3R1MH>;

fn main() {
    coremem::init_logging();
    coremem::init_debug();
    let um = |n| n as f32 * 1e-6;
    let ns = |n| n as f32 * 1e-9;
    let feat_size = um(10);

    let s_major = um(160);
    let s_minor = um(30);
    let io_major = um(80);
    let io_minor = um(30);
    let coupling_major = um(130);
    let coupling_minor = um(30);

    let sx = |n| um((n+1) * 400);
    let sy = um(400);
    let sz = um(280);
    let couplingx = |n| sx(n) - um(200);

    let sim_bounds = Meters::new(sx(4), sy * 2.0, sz * 2.0);
    let sim_padding = Meters::new(um(80), um(80), um(80));
    let duration = Seconds(ns(1));

    let drive0 = Torus::new_xz(Meters::new(sx(0) - s_major, sy, sz), io_major, io_minor);
    let sense3 = Torus::new_xz(Meters::new(sx(3) + s_major, sy, sz), io_major, io_minor);
    let ctl = |n| Torus::new_yz(Meters::new(sx(n), sy + s_major, sz), io_major, io_minor);
    let s = |n| Torus::new_xy(Meters::new(sx(n), sy, sz), s_major, s_minor);
    let coupling = |n| Torus::new_xz(Meters::new(couplingx(n), sy, sz), coupling_major, coupling_minor);

    let wire_mat = IsomorphicConductor::new(1e6);
    let ferro_mat = wire_mat;
    // let ferro_mat = Ferroxcube3R1MH::new(); // uncomment when ready to simulate for real

    let mut driver: SpirvDriver<Mat> = Driver::new_spirv(sim_bounds, feat_size);
    driver.add_classical_boundary(sim_padding);
    driver.fill_region(&drive0, wire_mat);
    driver.fill_region(&sense3, wire_mat);
    for core in 0..4 {
        driver.fill_region(&s(core), ferro_mat);
        driver.fill_region(&ctl(core), wire_mat);
        if core != 0 {
            driver.fill_region(&coupling(core), wire_mat);
        }
    }

    let prefix = "out/applications/multi_core_inverter/0/";
    let _ = std::fs::create_dir_all(&prefix);
    // driver.add_state_file(&*format!("{}state.bc", prefix), 9600);
    driver.add_serializer_renderer(&*format!("{}frame-", prefix), 36000, None);
    driver.add_csv_renderer(&*format!("{}meas.csv", prefix), 400, None);
    driver.set_steps_per_stim(200);

    driver.step_until(duration);
}
