use coremem::*;
use coremem::geom::*;
use coremem::real::{Real as _, ToFloat as _};
use coremem::sim::legacy::mat::Static;
use coremem::stim::AbstractStimulus;
use coremem::types::vec::Vec3;
use rand::rngs::StdRng;
use rand::{Rng as _, SeedableRng as _};

fn energy<R: Region>(s: &dyn SampleableSim, reg: &R) -> f32 {
    let e = f64::half() * s.map_sum_over_enumerated(reg, |_pos: Index, cell| {
        cell.e().mag_sq().to_f64()
    });
    e.cast()
}

fn energy_now_and_then<R: Region>(state: &mut StaticSim, reg: &R, frames: u32) -> (f32, f32) {
    let energy_0 = energy(state, reg);
    for _ in 0..frames {
        state.step();
    }
    let energy_1 = energy(state, reg);
    (energy_0, energy_1)
}

struct PmlStim<F> {
    /// Maps index -> (stim vector, stim frequency)
    f: F,
    t_end: f32,
    feat_size: f32,
}
impl<F: Fn(Index) -> (Vec3<f32>, f32) + Sync> AbstractStimulus for PmlStim<F> {
    fn at(&self, t_sec: f32, pos: Meters) -> (Vec3<f32>, Vec3<f32>) {
        let angle = t_sec/self.t_end*f32::two_pi();
        let gate = 0.5*(1.0 - angle.cos());
        let (e, hz) = (self.f)(pos.to_index(self.feat_size));
        let sig_angle = angle*hz;
        let sig = sig_angle.sin();
        (e*gate*sig, Vec3::zero())
    }
}

/// Apply some stimulus, and then let it decay and measure the ratio of energy left in the system
fn apply_stim_full_interior<F>(state: &mut StaticSim, frames: u32, f: F)
where F: Fn(Index) -> (Vec3<f32>, f32) + Sync // returns (E vector, omega)
{
    let stim = PmlStim {
        f,
        t_end: (frames as f32) * state.timestep(),
        feat_size: state.feature_size(),
    };
    for _t in 0..frames {
        state.apply_stimulus(&stim);
        state.step();
    }
}

fn apply_stim_over_region<R, F>(state: &mut StaticSim, frames: u32, reg: R, f: F)
where
    R: Region,
    F: Fn(Index) -> (Vec3<f32>, f32) + Sync,
{
    let feat = state.feature_size();
    apply_stim_full_interior(state, frames, |idx| {
        if reg.contains(idx.to_meters(feat)) {
            f(idx)
        } else {
            (Vec3::zero(), 0.0)
        }
    });
}

/// Stimulate each point in the region with a pseudorandom (but predictable) e wave
fn apply_chaotic_stim_over_region<R: Region>(state: &mut StaticSim, frames: u32, interior: R) {
    apply_stim_over_region(state, frames, interior, |idx| {
        let seed = (idx.x() as u64) ^ ((idx.y() as u64) << 16) ^ ((idx.z() as u64) << 32);
        let mut rng = StdRng::seed_from_u64(seed);
        let dir = Vec3::new(
            rng.gen_range(-1.0..1.0),
            rng.gen_range(-1.0..1.0),
            rng.gen_range(-1.0..1.0),
        );
        // XXX only works if it's a whole number. I suppose this makes sense though, as
        // other numbers would create higher harmonics when gated.
        let hz = rng.gen_range(0..=2);
        (dir, hz as _)
    })
}

fn chaotic_pml_test(state: &mut StaticSim, boundary: u32, padding: u32, frames: u32) -> f32 {
    let feat = state.feature_size();
    {
        let upper_left_idx = Index::unit()*padding;
        let lower_right_idx = state.size() - Index::unit() - upper_left_idx;
        let interior = Cube::new(upper_left_idx.to_meters(feat), lower_right_idx.to_meters(feat));
        apply_chaotic_stim_over_region(state, frames, interior);
    }

    let upper_left_idx = Index::unit()*boundary;
    let lower_right_idx = state.size() - Index::unit() - upper_left_idx;
    let sim_region = Cube::new(upper_left_idx.to_meters(feat), lower_right_idx.to_meters(feat));

    let (energy_0, energy_1) = energy_now_and_then(state, &sim_region, frames);
    // println!("Energy: {}/{}", energy_1, energy_0);
    energy_1/energy_0
}

#[allow(unused)]
fn state_for_pml(size: Index, boundary: Index, feat_size: f32, sc_coeff: f32, cond_coeff: f32, pow: f32) -> StaticSim {
    let mut state = StaticSim::new(size, feat_size);
    let timestep = state.timestep();
    state.fill_boundary_using(boundary, |boundary_ness| {
        let b = boundary_ness.elem_pow(pow);
        let coord_stretch = b * sc_coeff / timestep;
        let conductivity = Vec3::unit() * (b.mag() * cond_coeff / timestep);
        unimplemented!();
        Static::default()
        // Static {
        //     // TODO PML coord_stretch,
        //     conductivity,
        //     pml: Some((PmlState::new(), PmlParameters::new(coord_stretch))),
        //     ..Default::default()
        // }
    });
    state
}

fn main() {
    // Explanation here (slide 63): https://empossible.net/wp-content/uploads/2020/01/Lecture-The-Perfectly-Matched-Layer.pdf
    // Claims that eps0/delta_t * n^3 is a good way to activate the PML layer

    for pow in vec![3.0] {
    for sc_coeff in vec![
        // 0.1*consts::EPS0,
        // 0.5*consts::EPS0,
        // 1e0*consts::EPS0,
        // 1e1*consts::EPS0,
        // 1e2*consts::EPS0,
        // 1e3*consts::EPS0,
        // 1e6*consts::EPS0,
        // 1e9*consts::EPS0,
        // 1e12*consts::EPS0,
        // 1e15*consts::EPS0,
        // 1e18*consts::EPS0,
        // 1e21*consts::EPS0,
        // 0.01,
        // 0.03,
        // 0.05,
        // 0.07,
        // 0.08,
        // 0.09,
        // 0.1,
        // 0.15,
        // 0.2,
        // 0.25,
        // 0.3,
        // 0.4,
        // 0.5,
        // 0.6,
        // 0.7,
        // 0.8,
        // 0.9,
        //0.95,
        // 1.0,
        // 1.5,
        // 2.0,
        // 3.0,
        // 5.0,
        // 10.0,
        // 100.0,
        // 1000.0,
        //0.0,
        0.5,
    ] {
    //for cond_coeff in vec![0.0, 1.0, 1e3, 1e6, 1e9] {
    for cond_coeff in vec![0.0, 0.5*f32::eps0()] {
        for frames in vec![400, 1200] {
            for pml_width in vec![1, 2, 4, 8, 16] {
                for feat_size in vec![1e-6] {
                    let size = Index::unit()*121;
                    let sim_inset = 40;
                    let boundary = Index::unit()*pml_width;
                    let mut state = state_for_pml(size, boundary, feat_size, sc_coeff, cond_coeff, pow);
                    let ratio = chaotic_pml_test(&mut state, pml_width, sim_inset, frames);
                    println!("{},{} (pow={}, f={}, width={}, frames={}): {}", sc_coeff, cond_coeff, pow, feat_size, pml_width, frames, ratio);
                }
            }
        }
    }
    }
    }
    // Conclusions:
    // * if coordinate stretching is divided by time_step, then the absorption is independent of
    //   feature size.
    // For 240 frames, PML width of 20, size=121, sim_inset = 40, cubic onset:
    // * coefficients between [0.4, 1.0] are all within 30% of each other
    // * coefficients between [0.5, 1.0] are all within 20% of each other
    // * coefficients > 1 show instability (i.e. instable if coord_stretch > 1 / timestep);
    // * absorption generally increases as the coefficient approaches 1.0 from the left.
    //   This begins to reverse at least by 0.98
    // For 2400 frames, instability starts somewhere between 0.7 and 0.8
    //   0.5 (f=0.000001, width=20, frames=2400): 0.0006807010986857421
    //   0.6 (f=0.000001, width=20, frames=2400): 0.0005800974138738364
    //   0.7 (f=0.000001, width=20, frames=2400): 0.0005092274390086559
    //   0.8 (f=0.000001, width=20, frames=2400): 7326609898580109000000
    //   These numbers use hz = rng.gen_range(0..=5), with 20 frames of excitation and 20 steps
    //     between source and boundary
    //
    // Reducing hz to 0..=2 achieves much better perf (pow=3.0):
    //   0.5 (f=0.000001, width=20, frames=2400): 0.0000007135657380376141
    //   0.6 (f=0.000001, width=20, frames=2400): 0.0000006561934691721446
    //   0.7 (f=0.000001, width=20, frames=2400): 0.0000006099334150762627
    //   0.8 (f=0.000001, width=20, frames=2400): 2180868728529433400000
    // Quadratic (pow=2.0):
    //   0.4 (pow=2, f=0.000001, width=20, frames=2400): 0.0000006839243836328471
    //   0.5 (pow=2, f=0.000001, width=20, frames=2400): 0.0000006189051555210391
    //   0.6 (pow=2, f=0.000001, width=20, frames=2400): 0.9331062414894028
    //   0.7 (pow=2, f=0.000001, width=20, frames=2400): 222818414359827930000000000000000000000000000000000000000000000000000000000000000000000
    // Possibly better efficacy, but also less stable
    // Linear (pow=1.0) is 100% unstable for at least coeff >= 0.5
    // Subcubic (pow=2.5):
    //   0.4 (pow=2.5, f=0.000001, width=20, frames=2400): 0.0000007333781495449756
    //   0.5 (pow=2.5, f=0.000001, width=20, frames=2400): 0.0000006632381361190291
    //   0.6 (pow=2.5, f=0.000001, width=20, frames=2400): 0.0000006089982278171636
    //   0.7 (pow=2.5, f=0.000001, width=20, frames=2400): 7853952559.189004
    // Superlinear (pow=1.5):
    //   0.4 (pow=1.5, f=0.000001, width=20, frames=2400): 0.0000006502259131444893
    //   0.5 (pow=1.5, f=0.000001, width=20, frames=2400): 0.0000005956833662247809
    //   0.6 (pow=1.5, f=0.000001, width=20, frames=2400): 175615904583581400000000000000000000000000000000000000000000000000000000000000000000
    // pow=4.0:
    //   0.4 (pow=4, f=0.000001, width=20, frames=2400): 0.0000009105207758718423
    //   0.5 (pow=4, f=0.000001, width=20, frames=2400): 0.0000008169331995553951
    //   0.6 (pow=4, f=0.000001, width=20, frames=2400): 0.0000007525813705681166
    //   0.7 (pow=4, f=0.000001, width=20, frames=2400): 0.0000007019565296008103
    //   0.8 (pow=4, f=0.000001, width=20, frames=2400): 0.0000006600546205898854
    //   0.9 (pow=4, f=0.000001, width=20, frames=2400): 0.0000006246047577819345
    //   1 (pow=4, f=0.000001, width=20, frames=2400): 5437370254206590000000000000000000000000000000000
    // pow=3.5:
    //   0.4 (pow=3.5, f=0.000001, width=20, frames=2400): 0.0000008485707012112478
    //   0.5 (pow=3.5, f=0.000001, width=20, frames=2400): 0.0000007653501316913873
    //   0.6 (pow=3.5, f=0.000001, width=20, frames=2400): 0.0000007047188963430957
    //   0.7 (pow=3.5, f=0.000001, width=20, frames=2400): 0.0000006561795324449849
    //   0.8 (pow=3.5, f=0.000001, width=20, frames=2400): 0.0000006159621438553542
    //   0.9 (pow=3.5, f=0.000001, width=20, frames=2400): 18913062838424785000000000000000000
    // So generally lower powers = more absorbtion (makes sense: higher average coordinate
    // stretching), but is more sensitive to error. All powers > 1.5 are stable at coeff=0.5.
    // We don't know that much about reflection from just this data.
    // Running over a smaller timeframe should give some suggestion about reflection
    //   0.4 (pow=2, f=0.000001, width=20, frames=120): 0.2870021971493659
    //   0.5 (pow=2, f=0.000001, width=20, frames=120): 0.2647114916517679 **
    //   0.6 (pow=2, f=0.000001, width=20, frames=120): 0.24807341265362437
    //   0.7 (pow=2, f=0.000001, width=20, frames=120): 0.2350081813369852
    //   0.8 (pow=2, f=0.000001, width=20, frames=120): 0.22438565227661014
    //   0.9 (pow=2, f=0.000001, width=20, frames=120): 0.21552516261496907
    //   1.0 (pow=2, f=0.000001, width=20, frames=120): 0.20798702383197107
    //   1.5 (pow=2, f=0.000001, width=20, frames=120): 0.39444643124947426
    //   0.4 (pow=3, f=0.000001, width=20, frames=120): 0.3596522778473902
    //   0.5 (pow=3, f=0.000001, width=20, frames=120): 0.33718409856439474
    //   0.6 (pow=3, f=0.000001, width=20, frames=120): 0.32025414395335816
    //   0.7 (pow=3, f=0.000001, width=20, frames=120): 0.3067653188201421 **
    //   0.8 (pow=3, f=0.000001, width=20, frames=120): 0.2956250635204507
    //   0.9 (pow=3, f=0.000001, width=20, frames=120): 0.2861895432992242
    //   1.0 (pow=3, f=0.000001, width=20, frames=120): 0.27804573152917056
    //   1.5 (pow=3, f=0.000001, width=20, frames=120): 0.24950413937515897
    //   0.4 (pow=4, f=0.000001, width=20, frames=120): 0.41043160705678194
    //   0.5 (pow=4, f=0.000001, width=20, frames=120): 0.38844802581052806
    //   0.6 (pow=4, f=0.000001, width=20, frames=120): 0.3721037027266112
    //   0.7 (pow=4, f=0.000001, width=20, frames=120): 0.35912262947037576
    //   0.8 (pow=4, f=0.000001, width=20, frames=120): 0.34837556188585944
    //   0.9 (pow=4, f=0.000001, width=20, frames=120): 0.33922665697709947 **
    //   1 (pow=4, f=0.000001, width=20, frames=120): 0.33128127577349487
    //   1.5 (pow=4, f=0.000001, width=20, frames=120): 0.30260541581006584
    // Even smaller timeframe:
    //   0.5 (pow=2, f=0.000001, width=20, frames=5): 1.0109467232603757
    //   0.5 (pow=2, f=0.000001, width=20, frames=10): 1.0117049984675424
    //   0.5 (pow=2, f=0.000001, width=20, frames=20): 1.015054295037722
    //   0.5 (pow=2, f=0.000001, width=20, frames=30): 1.016127315217072
    //   0.5 (pow=2, f=0.000001, width=20, frames=40): 1.0125450428623481
    //   0.5 (pow=2, f=0.000001, width=20, frames=60): 0.9520114341681021
    //   0.5 (pow=2, f=0.000001, width=20, frames=80): 0.7617289087518861
    //   0.5 (pow=2, f=0.000001, width=20, frames=120): 0.2647114916517679
    //   0.5 (pow=2, f=0.000001, width=20, frames=160): 0.018698432021826004
    //   0.5 (pow=2, f=0.000001, width=20, frames=200): 0.001053437888276037
    //   0.5 (pow=3, f=0.000001, width=20, frames=5): 1.0109467232603757
    //   0.5 (pow=3, f=0.000001, width=20, frames=10): 1.0117049984674478
    //   0.5 (pow=3, f=0.000001, width=20, frames=20): 1.0150542142468075
    //   0.5 (pow=3, f=0.000001, width=20, frames=30): 1.0161755433699602
    //   0.5 (pow=3, f=0.000001, width=20, frames=40): 1.0140118995944478
    //   0.5 (pow=3, f=0.000001, width=20, frames=60): 0.9828724159361404
    //   0.5 (pow=3, f=0.000001, width=20, frames=80): 0.8304056107875255
    //   0.5 (pow=3, f=0.000001, width=20, frames=120): 0.33718409856439474
    //   0.5 (pow=3, f=0.000001, width=20, frames=160): 0.033906877503334515
    //   0.5 (pow=3, f=0.000001, width=20, frames=200): 0.0016075121270576818
    //   0.5 (pow=4, f=0.000001, width=20, frames=5): 1.0109467232603757
    //   0.5 (pow=4, f=0.000001, width=20, frames=10): 1.0117049984674424
    //   0.5 (pow=4, f=0.000001, width=20, frames=20): 1.015054203569572
    //   0.5 (pow=4, f=0.000001, width=20, frames=30): 1.0161827617273314
    //   0.5 (pow=4, f=0.000001, width=20, frames=40): 1.014355067000562
    //   0.5 (pow=4, f=0.000001, width=20, frames=60): 0.9971421502814346
    //   0.5 (pow=4, f=0.000001, width=20, frames=80): 0.8718457443603458
    //   0.5 (pow=4, f=0.000001, width=20, frames=120): 0.38844802581052806
    //   0.5 (pow=4, f=0.000001, width=20, frames=160): 0.04879003869310861
    //   0.5 (pow=4, f=0.000001, width=20, frames=200): 0.0025039595751290534
    //   That the numbers are all the same for t < 20 is not encouraging.
    //     Could be because of the courant number 0.577? No energy has reached the border yet?
    // Why don't we query just the energy in the sim region -- not the boundary?
    // After filtering to measure energy only in the sim region (note that some energy outside the
    // sim region COULD be reflected back into the sim region later):
    //   0.5 (pow=2, f=0.000001, width=20, frames=5): 1.0109467232603757
    //   0.5 (pow=2, f=0.000001, width=20, frames=10): 1.0117049984565956
    //   0.5 (pow=2, f=0.000001, width=20, frames=20): 1.0150051760172862
    //   0.5 (pow=2, f=0.000001, width=20, frames=30): 1.0103422463150569
    //   0.5 (pow=2, f=0.000001, width=20, frames=40): 0.9534926741875023
    //   0.5 (pow=2, f=0.000001, width=20, frames=60): 0.7052236436305864
    //   0.5 (pow=2, f=0.000001, width=20, frames=80): 0.42366103692252166
    //   0.5 (pow=2, f=0.000001, width=20, frames=120): 0.047117482623349645
    //   0.5 (pow=2, f=0.000001, width=20, frames=160): 0.0012788501490854916
    //   0.5 (pow=2, f=0.000001, width=20, frames=200): 0.00025052555946608536
    //   0.5 (pow=2, f=0.000001, width=20, frames=300): 0.000026163786700826282
    //   0.5 (pow=2, f=0.000001, width=20, frames=400): 0.000009482696335385068
    //   0.5 (pow=2, f=0.000001, width=20, frames=600): 0.0000034339203626681873
    //   0.5 (pow=2, f=0.000001, width=20, frames=800): 0.0000016501593549063537
    //   0.5 (pow=2, f=0.000001, width=20, frames=1200): 0.0000007097133927819365
    //   0.5 (pow=2, f=0.000001, width=20, frames=1600): 0.0000004990355277480898
    //   0.5 (pow=2, f=0.000001, width=20, frames=2400): 0.00000026541631012862064
    //   0.5 (pow=3, f=0.000001, width=20, frames=5): 1.0109467232603757
    //   0.5 (pow=3, f=0.000001, width=20, frames=10): 1.0117049984564943
    //   0.5 (pow=3, f=0.000001, width=20, frames=20): 1.015005103609574
    //   0.5 (pow=3, f=0.000001, width=20, frames=30): 1.010347173332254
    //   0.5 (pow=3, f=0.000001, width=20, frames=40): 0.9534945122673749
    //   0.5 (pow=3, f=0.000001, width=20, frames=60): 0.7052084303225945
    //   0.5 (pow=3, f=0.000001, width=20, frames=80): 0.42365590263359737
    //   0.5 (pow=3, f=0.000001, width=20, frames=120): 0.04711930602404218
    //   0.5 (pow=3, f=0.000001, width=20, frames=160): 0.0013058690825149086
    //   0.5 (pow=3, f=0.000001, width=20, frames=200): 0.00027124952640186995
    //   0.5 (pow=3, f=0.000001, width=20, frames=300): 0.00003363606115185493
    //   0.5 (pow=3, f=0.000001, width=20, frames=400): 0.000012979216745112146
    //   0.5 (pow=3, f=0.000001, width=20, frames=600): 0.000004790890253991059
    //   0.5 (pow=3, f=0.000001, width=20, frames=800): 0.0000018967555372300478
    //   0.5 (pow=3, f=0.000001, width=20, frames=1200): 0.000000785458387440694
    //   0.5 (pow=3, f=0.000001, width=20, frames=1600): 0.0000005657588895614973
    //   0.5 (pow=3, f=0.000001, width=20, frames=2400): 0.0000002931443936827707
    //   0.5 (pow=4, f=0.000001, width=20, frames=5): 1.0109467232603757
    //   0.5 (pow=4, f=0.000001, width=20, frames=10): 1.011704998456489
    //   0.5 (pow=4, f=0.000001, width=20, frames=20): 1.0150050988039427
    //   0.5 (pow=4, f=0.000001, width=20, frames=30): 1.0103477261980267
    //   0.5 (pow=4, f=0.000001, width=20, frames=40): 0.9534933477754826
    //   0.5 (pow=4, f=0.000001, width=20, frames=60): 0.7052093991486525
    //   0.5 (pow=4, f=0.000001, width=20, frames=80): 0.4236543609610083
    //   0.5 (pow=4, f=0.000001, width=20, frames=120): 0.047143352151935734
    //   0.5 (pow=4, f=0.000001, width=20, frames=160): 0.0014253338264054373
    //   0.5 (pow=4, f=0.000001, width=20, frames=200): 0.0003719236701949888
    //   0.5 (pow=4, f=0.000001, width=20, frames=300): 0.0000655709065994179
    //   0.5 (pow=4, f=0.000001, width=20, frames=400): 0.000019428436023699366
    //   0.5 (pow=4, f=0.000001, width=20, frames=600): 0.000006787101433521501
    //   0.5 (pow=4, f=0.000001, width=20, frames=800): 0.000002352253311002263
    //   0.5 (pow=4, f=0.000001, width=20, frames=1200): 0.0000008945009475225972
    //   0.5 (pow=4, f=0.000001, width=20, frames=1600): 0.0000006298015155592009
    //   0.5 (pow=4, f=0.000001, width=20, frames=2400): 0.0000003110877898243663
    //   pow=2 and pow=3 look nearly interchangeable in perf, with pow=4 slightly worse.
    //   Given higher stability for pow=3, it pushes me in that direction
    // What about ordinary conductivity?
    //   Uniaxial conductors do nothing
    // non-axial conductors (measured over full volume -- not just sim volume)
    //   0,0.000000000000001 (pow=2, f=0.000001, width=20, frames=20): 1.0150542017357447
    //   0,0.000000000000001 (pow=2, f=0.000001, width=20, frames=40): 1.0144968836020802
    //   0,0.000000000000001 (pow=2, f=0.000001, width=20, frames=80): 1.001249693707569
    //   0,0.000000000000001 (pow=2, f=0.000001, width=20, frames=120): 0.9703627391410313
    //   0,0.000000000000001 (pow=2, f=0.000001, width=20, frames=240): 0.9839347439326915
    //   0,0.000000000000001 (pow=2, f=0.000001, width=20, frames=400): 0.9710971333858839
    //   0,0.000000000001 (pow=2, f=0.000001, width=20, frames=20): 1.015054152058065
    //   0,0.000000000001 (pow=2, f=0.000001, width=20, frames=40): 1.0139965206685178
    //   0,0.000000000001 (pow=2, f=0.000001, width=20, frames=80): 0.8611234324272511
    //   0,0.000000000001 (pow=2, f=0.000001, width=20, frames=120): 0.38724876605299674
    //   0,0.000000000001 (pow=2, f=0.000001, width=20, frames=240): 0.010452374686360136
    //   0,0.000000000001 (pow=2, f=0.000001, width=20, frames=400): 0.0006325928742141608
    //   0,0.000000001 (pow=2, f=0.000001, width=20, frames=20): 1.0150333781336882
    //   0,0.000000001 (pow=2, f=0.000001, width=20, frames=40): 0.9834360538048873
    //   0,0.000000001 (pow=2, f=0.000001, width=20, frames=80): 0.5072413415421974
    //   0,0.000000001 (pow=2, f=0.000001, width=20, frames=120): 0.10345998021237998
    //   0,0.000000001 (pow=2, f=0.000001, width=20, frames=240): 0.007519161562708659
    //   0,0.000000001 (pow=2, f=0.000001, width=20, frames=400): 0.0006070229440826888
    //   0,0.000001 (pow=2, f=0.000001, width=20, frames=20): 1.0150056763216164
    //   0,0.000001 (pow=2, f=0.000001, width=20, frames=40): 1.0123907067781637
    //   0,0.000001 (pow=2, f=0.000001, width=20, frames=80): 1.0095377475794036
    //   0,0.000001 (pow=2, f=0.000001, width=20, frames=120): 1.0027226600640793
    //   0,0.000001 (pow=2, f=0.000001, width=20, frames=240): 1.002880617201309
    //   0,0.000001 (pow=2, f=0.000001, width=20, frames=400): 0.9817776859704778
    //   0,0.001 (pow=2, f=0.000001, width=20, frames=20): 1.0150056260987463
    //   0,0.001 (pow=2, f=0.000001, width=20, frames=40): 1.0127202910529511
    //   0,0.001 (pow=2, f=0.000001, width=20, frames=80): 1.0141889645709163
    //   0,0.001 (pow=2, f=0.000001, width=20, frames=120): 1.012712629869813
    //   0,0.001 (pow=2, f=0.000001, width=20, frames=240): 1.0225762750160987
    //   0,0.001 (pow=2, f=0.000001, width=20, frames=400): 1.0158169138074233
    //   0,1 (pow=2, f=0.000001, width=20, frames=20): 1.0150056260488272
    //   0,1 (pow=2, f=0.000001, width=20, frames=40): 1.0127206223077903
    //   0,1 (pow=2, f=0.000001, width=20, frames=80): 1.01419363999915
    //   0,1 (pow=2, f=0.000001, width=20, frames=120): 1.0127226953239867
    //   0,1 (pow=2, f=0.000001, width=20, frames=240): 1.0225963209694455
    //   0,1 (pow=2, f=0.000001, width=20, frames=400): 1.0158520285873147
    //   0,0.000000000000001 (pow=3, f=0.000001, width=20, frames=20): 1.0150542017810857
    //   0,0.000000000000001 (pow=3, f=0.000001, width=20, frames=40): 1.014497284230248
    //   0,0.000000000000001 (pow=3, f=0.000001, width=20, frames=80): 1.0013349322815948
    //   0,0.000000000000001 (pow=3, f=0.000001, width=20, frames=120): 0.9707653838346726
    //   0,0.000000000000001 (pow=3, f=0.000001, width=20, frames=240): 0.98542720202683
    //   0,0.000000000000001 (pow=3, f=0.000001, width=20, frames=400): 0.9736668760490382
    //   0,0.000000000001 (pow=3, f=0.000001, width=20, frames=20): 1.0150541972999592
    //   0,0.000000000001 (pow=3, f=0.000001, width=20, frames=40): 1.0143834319294165
    //   0,0.000000000001 (pow=3, f=0.000001, width=20, frames=80): 0.9087702844329181
    //   0,0.000000000001 (pow=3, f=0.000001, width=20, frames=120): 0.4641021861485614
    //   0,0.000000000001 (pow=3, f=0.000001, width=20, frames=240): 0.025542565246607578
    //   0,0.000000000001 (pow=3, f=0.000001, width=20, frames=400): 0.0024678463432975324
    //   0,0.000000001 (pow=3, f=0.000001, width=20, frames=20): 1.0150507306076173
    //   0,0.000000001 (pow=3, f=0.000001, width=20, frames=40): 0.9973368290451762
    //   0,0.000000001 (pow=3, f=0.000001, width=20, frames=80): 0.5359561255382069
    //   0,0.000000001 (pow=3, f=0.000001, width=20, frames=120): 0.10107229161449789
    //   0,0.000000001 (pow=3, f=0.000001, width=20, frames=240): 0.0018726822600725027
    //   0,0.000000001 (pow=3, f=0.000001, width=20, frames=400): 0.0001395247365920271
    //   0,0.000001 (pow=3, f=0.000001, width=20, frames=20): 1.015006719517991
    //   0,0.000001 (pow=3, f=0.000001, width=20, frames=40): 1.0067208901320701
    //   0,0.000001 (pow=3, f=0.000001, width=20, frames=80): 0.9295573543881088
    //   0,0.000001 (pow=3, f=0.000001, width=20, frames=120): 0.8372562924822498
    //   0,0.000001 (pow=3, f=0.000001, width=20, frames=240): 0.7109091146305759
    //   0,0.000001 (pow=3, f=0.000001, width=20, frames=400): 0.5405933200528408
    //   0,0.001 (pow=3, f=0.000001, width=20, frames=20): 1.015005627048278
    //   0,0.001 (pow=3, f=0.000001, width=20, frames=40): 1.0127139915480314
    //   0,0.001 (pow=3, f=0.000001, width=20, frames=80): 1.0141000517606538
    //   0,0.001 (pow=3, f=0.000001, width=20, frames=120): 1.0125212238134818
    //   0,0.001 (pow=3, f=0.000001, width=20, frames=240): 1.0221951620616994
    //   0,0.001 (pow=3, f=0.000001, width=20, frames=400): 1.0151495300965783
    //   0,1 (pow=3, f=0.000001, width=20, frames=20): 1.0150056260497768
    //   0,1 (pow=3, f=0.000001, width=20, frames=40): 1.012720616007617
    //   0,1 (pow=3, f=0.000001, width=20, frames=80): 1.0141935510766382
    //   0,1 (pow=3, f=0.000001, width=20, frames=120): 1.0127225038875007
    //   0,1 (pow=3, f=0.000001, width=20, frames=240): 1.0225959397081388
    //   0,1 (pow=3, f=0.000001, width=20, frames=400): 1.0158513607157937
    //   It appears here that the cubic roll-off is best, and very low conductivities are required.
    //
    // Isotropic conductor, energy measured only within the sim area:
    //   The optimized case (0.5*EPS0/timestep * x^3) is almost IDENTICAL to the
    //   optimized stretched-coordinate version.
    //   0,0.000000000000001 (pow=2, f=0.000001, width=20, frames=120): 0.07433102240945513
    //   0,0.000000000000001 (pow=2, f=0.000001, width=20, frames=400): 0.20431918328907037
    //   0,0.000000000001 (pow=2, f=0.000001, width=20, frames=120): 0.04851013262210713
    //   0,0.000000000001 (pow=2, f=0.000001, width=20, frames=400): 0.00018366232272528987
    //   0,0.0000000000044270939064065 (pow=2, f=0.000001, width=20, frames=120): 0.047101055930619994
    //   0,0.0000000000044270939064065 (pow=2, f=0.000001, width=20, frames=400): 0.000007399479477711689
    //   0,0.000000000008854187812813 (pow=2, f=0.000001, width=20, frames=120): 0.047132676278627536
    //   0,0.000000000008854187812813 (pow=2, f=0.000001, width=20, frames=400): 0.000008105989741898135
    //   0,0.00000000001 (pow=2, f=0.000001, width=20, frames=120): 0.047145523858393434
    //   0,0.00000000001 (pow=2, f=0.000001, width=20, frames=400): 0.000008327266269591764
    //   0,0.0000000001 (pow=2, f=0.000001, width=20, frames=120): 0.05005311352870215
    //   0,0.0000000001 (pow=2, f=0.000001, width=20, frames=400): 0.00003665711687010239
    //   0,0.000000001 (pow=2, f=0.000001, width=20, frames=120): 0.0808038989604745
    //   0,0.000000001 (pow=2, f=0.000001, width=20, frames=400): 0.0005636469832522345
    //   0,0.00000001 (pow=2, f=0.000001, width=20, frames=120): 0.42090062694288544
    //   0,0.00000001 (pow=2, f=0.000001, width=20, frames=400): 0.07558396270665715
    //   0,0.000001 (pow=2, f=0.000001, width=20, frames=120): 0.9456301791158588
    //   0,0.000001 (pow=2, f=0.000001, width=20, frames=400): 0.9456217562930543
    //   0,0.001 (pow=2, f=0.000001, width=20, frames=120): 0.9549600788853474
    //   0,0.001 (pow=2, f=0.000001, width=20, frames=400): 0.9782784298162882
    //   0,1 (pow=2, f=0.000001, width=20, frames=120): 0.9549694776242036
    //   0,1 (pow=2, f=0.000001, width=20, frames=400): 0.9783121245194134
    //   0,0.000000000000001 (pow=3, f=0.000001, width=20, frames=120): 0.07435108171392021
    //   0,0.000000000000001 (pow=3, f=0.000001, width=20, frames=400): 0.20479144240774633
    //   0,0.000000000001 (pow=3, f=0.000001, width=20, frames=120): 0.050005777221700895
    //   0,0.000000000001 (pow=3, f=0.000001, width=20, frames=400): 0.0006548443359188922
    //   0,0.0000000000044270939064065 (pow=3, f=0.000001, width=20, frames=120): 0.0471097820852083
    //   0,0.0000000000044270939064065 (pow=3, f=0.000001, width=20, frames=400): 0.00000726683693079002
    //   0,0.000000000008854187812813 (pow=3, f=0.000001, width=20, frames=120): 0.04711795151894901
    //   0,0.000000000008854187812813 (pow=3, f=0.000001, width=20, frames=400): 0.000007527865420649209
    //   0,0.00000000001 (pow=3, f=0.000001, width=20, frames=120): 0.0471214615521427
    //   0,0.00000000001 (pow=3, f=0.000001, width=20, frames=400): 0.000007561968840324101
    //   0,0.0000000001 (pow=3, f=0.000001, width=20, frames=120): 0.04775494018209187
    //   0,0.0000000001 (pow=3, f=0.000001, width=20, frames=400): 0.00001655968038405813
    //   0,0.000000001 (pow=3, f=0.000001, width=20, frames=120): 0.05548225210925576
    //   0,0.000000001 (pow=3, f=0.000001, width=20, frames=400): 0.00011973516700797696
    //   0,0.00000001 (pow=3, f=0.000001, width=20, frames=120): 0.09859866928724702
    //   0,0.00000001 (pow=3, f=0.000001, width=20, frames=400): 0.0010788775694009376
    //   0,0.000001 (pow=3, f=0.000001, width=20, frames=120): 0.7905518369602526
    //   0,0.000001 (pow=3, f=0.000001, width=20, frames=400): 0.5217962235754597
    //   0,0.001 (pow=3, f=0.000001, width=20, frames=120): 0.9547813505519078
    //   0,0.001 (pow=3, f=0.000001, width=20, frames=400): 0.9776380394998598
    //   0,1 (pow=3, f=0.000001, width=20, frames=120): 0.9549692988681137
    //   0,1 (pow=3, f=0.000001, width=20, frames=400): 0.978311483657099
    // With both PML and conductor boundary:
    //   0.5,0.0000000000044270939064065 (pow=2, f=0.000001, width=20, frames=120): 0.04709454626708049
    //   0.5,0.0000000000044270939064065 (pow=2, f=0.000001, width=20, frames=400): 0.000007626399231684735
    //   0.5,0.0000000000044270939064065 (pow=2, f=0.000001, width=20, frames=1200): 0.0000008116632120088743
    //   0.5,0.0000000000044270939064065 (pow=2, f=0.000001, width=20, frames=2400): 0.00000032497678753300923
    //   0.5,0.0000000000044270939064065 (pow=3, f=0.000001, width=20, frames=120): 0.04709779630205149
    //   0.5,0.0000000000044270939064065 (pow=3, f=0.000001, width=20, frames=400): 0.000007132448872463483
    //   0.5,0.0000000000044270939064065 (pow=3, f=0.000001, width=20, frames=1200): 0.000000710451109532798
    //   0.5,0.0000000000044270939064065 (pow=3, f=0.000001, width=20, frames=2400): 0.0000002515070714149805
    //   This is basically no change from JUST PML or JUST conductors
    //   Maybe I should be trying to vary the width: maybe PML works more effectively for narrower
    //   boundaries than do conductors?
}
