use crate::geom::{Coord, Meters, OrdMeters};

use dyn_clone::{self, DynClone};
use rayon::prelude::*;
use serde::{Serialize, Deserialize};
use std::collections::BTreeSet;
use std::sync::Arc;

mod primitives;
pub use primitives::*;


#[typetag::serde(tag = "type")]
pub trait Region: Send + Sync + DynClone {
    fn contains(&self, p: Meters) -> bool;
}
dyn_clone::clone_trait_object!(Region);

pub fn and<T1: Region + 'static, T2: Region + 'static>(r1: T1, r2: T2) -> Intersection {
    Intersection::new().and(r1).and(r2)
}

pub fn and_not<T1: Region + 'static, T2: Region + 'static>(r1: T1, r2: T2) -> Intersection {
    and(r1, InvertedRegion::new(r2))
}

pub fn union<T1: Region + 'static, T2: Region + 'static>(r1: T1, r2: T2) -> Union {
    Union::new().with(r1).with(r2)
}

/// returns true if there's a path (via the cardinal directions) from p0 to p1 within this region.
pub fn is_connected<R: Region, C: Coord>(r: &R, p0: C, p1: C, feat_size: f32) -> bool {
    distance_to(r, p0, p1, feat_size).is_some()
}

/// returns the minimal distance to crawl from p0 to p1 (via movements along a cardinal direction)
pub fn distance_to<R: Region, C: Coord>(r: &R, p0: C, p1: C, feat_size: f32) -> Option<f32> {
    // breadth-first crawl, where `visited` contains everything at an earlier depth
    // and `active_layer` contains nodes which are a minimal of `active_hops` from p0.
    let mut visited = BTreeSet::new();
    let mut active_layer = BTreeSet::new();
    active_layer.insert(p0.to_index(feat_size));
    let mut active_hops = 0;

    let p1idx = p1.to_index(feat_size);

    while !active_layer.is_empty() {
        // we have to mark these as visited *before* crawling them,
        // else neighbors might be crawled twice.
        visited.append(&mut active_layer.clone());

        if visited.contains(&p1idx) {
            return Some(active_hops as f32 * feat_size);
        }

        // add the neighbors to the list of places to crawl
        active_layer = active_layer
            .into_par_iter()
            .flat_map(|loc| loc.cardinal_neighbors())
            .filter(|neighbor| {
                !visited.contains(&neighbor) && r.contains(neighbor.to_meters(feat_size))
            })
            .collect();
        active_hops += 1;
    }
    None // not found
}

/// Region describing the entire simulation space
#[derive(Copy, Clone, Serialize, Deserialize)]
pub struct WorldRegion;

#[typetag::serde]
impl Region for WorldRegion {
    fn contains(&self, _: Meters) -> bool {
        true
    }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct InvertedRegion(Box<dyn Region>);

impl InvertedRegion {
    pub fn new<R: Region + 'static>(r: R) -> Self {
        Self(Box::new(r))
    }
}

#[typetag::serde]
impl Region for InvertedRegion {
    fn contains(&self, p: Meters) -> bool {
        !self.0.contains(p)
    }
}

#[derive(Clone, Default, Serialize, Deserialize)]
pub struct Union(Vec<Box<dyn Region>>);

impl Union {
    pub fn new() -> Self {
        Self(Vec::new())
    }
    pub fn new_with<R: Region + 'static>(r: R) -> Self {
        Self::new().with(r)
    }
    pub fn with<R: Region + 'static>(self, r: R) -> Self {
        self.with_box(Box::new(r))
    }
    pub fn with_box(mut self, r: Box<dyn Region>) -> Self {
        self.0.push(r);
        self
    }
}

#[typetag::serde]
impl Region for Union {
    fn contains(&self, p: Meters) -> bool {
        self.0.iter().any(|r| r.contains(p))
    }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct Intersection(Vec<Box<dyn Region>>);


impl Intersection {
    pub fn new() -> Self {
        Self(Vec::new())
    }
    pub fn new_with<R: Region + 'static>(r: R) -> Self {
        Self::new().and(r)
    }
    pub fn and<R: Region + 'static>(self, r: R) -> Self {
        self.and_box(Box::new(r))
    }
    pub fn and_box(mut self, r: Box<dyn Region>) -> Self {
        self.0.push(r);
        self
    }
}

#[typetag::serde]
impl Region for Intersection {
    fn contains(&self, p: Meters) -> bool {
        self.0.iter().all(|r| r.contains(p))
    }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct Translate {
    inner: Box<dyn Region>,
    shift: Meters,
}

impl Translate {
    pub fn new<T: Region + 'static>(inner: T, shift: Meters) -> Self {
        Self { inner: Box::new(inner), shift }
    }
}

#[typetag::serde]
impl Region for Translate {
    fn contains(&self, p: Meters) -> bool {
        self.inner.contains(p - self.shift)
    }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct SwapXZ {
    inner: Box<dyn Region>,
}

impl SwapXZ {
    pub fn new<T: Region + 'static>(inner: T) -> Self {
        Self { inner: Box::new(inner) }
    }
}

#[typetag::serde]
impl Region for SwapXZ {
    fn contains(&self, p: Meters) -> bool {
        let p = Meters::new(p.z(), p.y(), p.z());
        self.inner.contains(p)
    }
}


#[derive(Clone, Serialize, Deserialize)]
pub struct SwapYZ {
    inner: Box<dyn Region>,
}

impl SwapYZ {
    pub fn new<T: Region + 'static>(inner: T) -> Self {
        Self { inner: Box::new(inner) }
    }
}

#[typetag::serde]
impl Region for SwapYZ {
    fn contains(&self, p: Meters) -> bool {
        let mapped = Meters::new(p.x(), p.z(), p.y());
        self.inner.contains(mapped)
    }
}

/// "wraps" the inner region into a donut shape.
/// this donut is traced out over the xy plane, leaving z untouched.
/// the center of the donut is (0, 0).
/// the resulting region is mapped onto the original region y=[0, y_max]. x is just the radius
/// so that (0, 0) is mapped to (0, 0), and (1, 0) is mapped to (1, 0) and (0, 1) is mapped to
/// (1, 0.5*y_max) and (-5, 0) is mapped to (5, 0.5*y_max).
#[derive(Clone, Serialize, Deserialize)]
pub struct Wrap {
    inner: Box<dyn Region>,
    y_max: f32,
    about: Meters,
}

impl Wrap {
    pub fn new<T: Region + 'static>(inner: T, y_max: f32) -> Self {
        Self::new_about(inner, y_max, Meters::new(0.0, 0.0, 0.0))
    }

    pub fn new_about<T: Region + 'static>(inner: T, y_max: f32, about: Meters) -> Self {
        Self { inner: Box::new(inner), y_max, about }
    }

    fn map(&self, p: Meters) -> Meters {
        let p_rel = p - self.about;
        let xy = p_rel.xy();
        let arg = xy.y().atan2(xy.x());
        let rev = arg / std::f32::consts::PI * self.y_max;
        Meters::new(xy.mag() + self.about.x(), rev, p.z())
    }
}

#[typetag::serde]
impl Region for Wrap {
    fn contains(&self, p: Meters) -> bool {
        self.inner.contains(self.map(p))
    }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct Dilate {
    inner: Box<dyn Region>,
    rad: f32,
    res: f32,
}

impl Dilate {
    pub fn new<T: Region + 'static>(inner: T, rad: f32, res: f32) -> Self {
        Self { inner: Box::new(inner), rad, res }
    }
}

#[typetag::serde]
impl Region for Dilate {
    fn contains(&self, p: Meters) -> bool {
        let rad_iters = (self.rad / self.res).ceil() as i32;
        let rad_range = -rad_iters..=rad_iters;
        let rad_sq = self.rad * self.rad;
        for zoff in rad_range.clone() {
            for yoff in rad_range.clone() {
                for xoff in rad_range.clone() {
                    let off = Meters::new(
                        xoff as f32 * self.res,
                        yoff as f32 * self.res,
                        zoff as f32 * self.res,
                    );
                    if off.mag_sq() <= rad_sq && self.inner.contains(p + off) {
                        return true;
                    }
                }
            }
        }
        false
    }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct Memoize {
    #[serde(skip)]
    lut: Arc<dashmap::DashMap<OrdMeters, bool>>,
    inner: Box<dyn Region>,
}

impl Memoize {
    pub fn new<R: Region + 'static>(inner: R) -> Self {
        Self {
            lut: Arc::new(dashmap::DashMap::new()),
            inner: Box::new(inner),
        }
    }
}

#[typetag::serde]
impl Region for Memoize {
    fn contains(&self, p: Meters) -> bool {
        *self.lut.entry(OrdMeters(p)).or_insert_with(|| self.inner.contains(p))
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use float_eq::assert_float_eq;

    fn assert_map(w: &Wrap, from: Meters, to: Meters) {
        let mapped = w.map(from);
        assert_float_eq!(mapped.x(), to.x(), abs <= 0.01);
        assert_float_eq!(mapped.y(), to.y(), abs <= 0.01);
        assert_float_eq!(mapped.z(), to.z(), abs <= 0.01);
    }

    #[test]
    fn wrap_basics() {
        let wrap = Wrap::new(Cube::default(), 10.0);

        // magnitude
        assert_map(&wrap, Meters::new(0.0, 0.0, 0.0), Meters::new(0.0, 0.0, 0.0));
        assert_map(&wrap, Meters::new(1.0, 0.0, 0.0), Meters::new(1.0, 0.0, 0.0));
        assert_map(&wrap, Meters::new(5.0, 0.0, 0.0), Meters::new(5.0, 0.0, 0.0));

        // angle
        assert_map(&wrap, Meters::new(0.0, 4.0, 0.0), Meters::new(4.0, 5.0, 0.0));
        assert_map(&wrap, Meters::new(0.0, -0.5, 0.0), Meters::new(0.5, -5.0, 0.0));
        assert_map(&wrap, Meters::new(1.0, 1.0, 0.0), Meters::new(1.41, 2.5, 0.0));
        assert_map(&wrap, Meters::new(-2.0, 0.0, 0.0), Meters::new(2.0, 10.0, 0.0));

        // z
        assert_map(&wrap, Meters::new(0.0, 0.0, 3.0), Meters::new(0.0, 0.0, 3.0));
        assert_map(&wrap, Meters::new(1.0, 1.0, -5.0), Meters::new(1.41, 2.5, -5.0));

    }

    #[test]
    fn wrap_about() {
        let wrap = Wrap::new_about(
            Cube::new(Meters::new(0.0, 0.0, 0.0), Meters::new(5.0, 5.0, 5.0)),
            10.0,
            Meters::new(-10.0, 0.0, -1000.0 /* irrelevant */),
        );

        assert_map(&wrap, Meters::new(1.0, 0.0, 0.0), Meters::new(1.0, 0.0, 0.0));
        assert_map(&wrap, Meters::new(-10.0, 10.0, 0.0), Meters::new(0.0, 5.0, 0.0));
        assert_map(&wrap, Meters::new(-20.0, 0.0, 0.0), Meters::new(0.0, 10.0, 0.0));

        // verify resulting x component
        assert_map(&wrap, Meters::new(-21.0, 0.0, 0.0), Meters::new(1.0, 10.0, 0.0));
        assert_map(&wrap, Meters::new(-19.0, 0.0, 0.0), Meters::new(-1.0, 10.0, 0.0));
        assert_map(&wrap, Meters::new(-10.0, 9.0, 0.0), Meters::new(-1.0, 5.0, 0.0));
        assert_map(&wrap, Meters::new(-10.0, -9.0, 0.0), Meters::new(-1.0, -5.0, 0.0));

        // z should be no-op
        assert_map(&wrap, Meters::new(-10.0, -9.0, 5.0), Meters::new(-1.0, -5.0, 5.0));
    }
}
