use super::Material;
use crate::geom::{Line2d, Polygon2d};
use crate::real::Real;
use crate::sim::legacy::{CellState, StepParametersMut};
use crate::types::vec::{Vec2, Vec3};

use lazy_static::lazy_static;
use log::trace;
use serde::{Serialize, Deserialize};
use std::any::{Any, TypeId};
use std::cmp::Ordering;
use std::collections::HashMap;
use std::sync::Mutex;

fn step_linear_ferro<R: Real>(m_mut: &mut Vec3<R>, mh_curve: &MHCurve<R>, context: &CellState<R>, delta_b: Vec3<R>) {
    trace!("step_b enter");
    let (h, m) = (context.h(), *m_mut);
    let target_hm = h + m + delta_b * R::mu0_inv();

    // TODO: this is probably not the best way to generalize a BH curve into 3d.
    let (_hx, mx) = mh_curve.move_to(
        h.x(),
        m.x(),
        target_hm.x(),
    );
    let (_hy, my) = mh_curve.move_to(
        h.y(),
        m.y(),
        target_hm.y(),
    );
    let (_hz, mz) = mh_curve.move_to(
        h.z(),
        m.z(),
        target_hm.z(),
    );

    *m_mut = Vec3::new(mx, my, mz);
    // let ret = Vec3::new(hx, hy, hz);
    trace!("step_b end");
}

/// M as a function of H
#[derive(Clone, PartialEq)]
struct MHCurve<R> {
    geom: Polygon2d<R>,
}

#[allow(unused)]
impl<R: Real> MHCurve<R> {
    /// Construct a M(H) curve from a sweep from M = 0 to Ms and back down to M = 0.
    /// The curve below M = 0 is derived by symmetry.
    fn new<R2: Real>(points: &[Vec2<R2>]) -> Self {
        let full_pts: Vec<_> =
            points.iter().cloned()
            .chain(points.iter().cloned().map(|p| -p))
            .map(|p| p.cast())
            .collect();

        Self {
            geom: Polygon2d::new(full_pts)
        }
    }

    fn from_bh<R2: Real>(points: &[(R2, R2)]) -> Self {
        let mh_points: Vec<_> = points.iter().cloned().map(|(h, b)| {
            Vec2::new(h, b / R2::mu0() - h)
        }).collect();

        Self::new(&*mh_points)
    }

    fn from_mh<R2: Real>(points: &[(R2, R2)]) -> Self {
        let mh_points: Vec<_> = points.iter().cloned().map(|(h, m)| {
            Vec2::new(h, m)
        }).collect();

        Self::new(&*mh_points)
    }

    /// Return (Hmax, Mmax)
    pub fn extremes(&self) -> Vec2<R> {
        Vec2::new(self.geom.max_x(), self.geom.max_y())
    }

    /// Moves (h, m) towards some location in the MH curve where H + M = target_hm.
    /// Returns `Ok((h, m))` if complete; `Err((h, m))` if there's more work to be done (call it
    /// again).
    fn step_toward(&self, h: R, m: R, target_hm: R) -> Result<Vec2<R>, Vec2<R>> {
        let is_ascending = match target_hm.partial_cmp(&(h + m)).unwrap_or_else(|| panic!("{} {}", h, m)) {
            Ordering::Greater => true,
            Ordering::Less => false,
            _ => return Ok(Vec2::new(h, m))
        };
        if (is_ascending && m == self.geom.max_y()) || (!is_ascending && m == self.geom.min_y()) {
            // Fully saturated. m is fixed, while h moves freely
            return Ok(Vec2::new(target_hm - m, m));
        }
        // Locate the segment which would contain the current point
        let mut segments = self.geom.segments();
        let active_segment = loop {
            let line = segments.next().unwrap_or_else(|| {
                panic!("failed to find segment for h:{}, m:{}, {:?}", h, m, self.geom.segments().collect::<Vec<_>>());
            });
            if line.contains_y(m) && line.is_ascending() == is_ascending {
                if line.contains_x(h) && line.distance_sq(Vec2::new(h, m)) < R::from_primitive(1.0e-6) {
                    // (h, m) resides on this line
                    break line;
                } else {
                    // need to move the point toward this line
                    let h_intercept = line.x(m);
                    break Line2d::new(Vec2::new(h, m), Vec2::new(h_intercept, m));
                }
            }
        };
        trace!("active segment: {:?}", active_segment);

        // Find some m(h) on the active_segment such that sum(h) = h + m(h) = target_hm
        let sum_h = active_segment + Line2d::new(Vec2::zero(), Vec2::unit());
        trace!("sum_h: {:?}", sum_h);
        let new_h = if sum_h.to().y() != sum_h.from().y() {
            sum_h.move_toward_y_unclamped(h, target_hm)
        } else {
            // avoid a division-by-zero.
            // We could be anywhere along this line, but we prefer the endpoint
            // so as to escape out of any permanent loops
            active_segment.to().x()
        };
        trace!("new_h: {}", new_h);

        if sum_h.contains_x(new_h) {
            // the segment contains a point with the target H+M
            Ok(active_segment.at_x(new_h))
        } else {
            // the segment doesn't contain the desired point: clamp and try the next segment
            Err(active_segment.clamp_by_x(new_h))
        }
    }
    fn move_to(&self, mut h: R, mut m: R, target_hm: R) -> (R, R) {
        let mut i = 0;
        loop {
            i += 1;
            match self.step_toward(h, m, target_hm) {
                Ok(v) => break (v.x(), v.y()),
                Err(v) => {
                    h = v.x();
                    m = v.y();
                },
            }
            if i % 2048 == 0 {
                panic!("unusually high iteration count without converging: {}. args: {}, {}, {}", i, h, m, target_hm);
            }
        }
    }
}

#[derive(Default, Copy, Clone, PartialEq, Serialize, Deserialize)]
pub struct Ferroxcube3R1<R> {
    m: Vec3<R>,
}

impl<R: Real> Ferroxcube3R1<R> {
    pub fn new() -> Self {
        Self::default()
    }
}

impl<R: Real> Ferroxcube3R1<R> {
    fn curve() -> &'static MHCurve<R> {
        lazy_static! {
            static ref CURVES: Mutex<HashMap<TypeId, Box<dyn Any + Send>>> = Mutex::new(HashMap::new());
        }
        let mut lock = CURVES.lock().unwrap();
        let curve = lock.entry(TypeId::of::<R>()).or_insert_with(|| {
            Box::new(MHCurve::<R>::from_bh(&[
                (  35.0, 0.0),
                (  50.0, 0.250),
                ( 100.0, 0.325),
                ( 200.0, 0.350),
                (1000.0, 0.390),
                // Falling
                ( 200.0, 0.360),
                ( 100.0, 0.345),
                (  50.0, 0.340),
                (   0.0, 0.325),
            ]))
        }).downcast_ref::<MHCurve<R>>().unwrap();
        unsafe { std::mem::transmute::<&MHCurve<R>, &'static MHCurve<R>>(curve) }
    }
}

impl<R: Real> Material<R> for Ferroxcube3R1<R> {
    fn step_b(&mut self, context: &CellState<R>, delta_b: Vec3<R>) {
        step_linear_ferro(&mut self.m, Self::curve(), context, delta_b)
    }
    fn m(&self) -> Vec3<R> {
        self.m
    }
    fn step_parameters_mut<'a>(&'a mut self) -> StepParametersMut<'a, R> {
        StepParametersMut::default().with_conductivity(Vec3::uniform(1e-3))
    }
}

/// Simple, square-loop ferrite
#[derive(Default, Copy, Clone, PartialEq, Serialize, Deserialize)]
pub struct MinimalSquare<R> {
    m: Vec3<R>,
}

impl<R: Real> MinimalSquare<R> {
    fn curve() -> &'static MHCurve<R> {
        lazy_static! {
            static ref CURVES: Mutex<HashMap<TypeId, Box<dyn Any + Send>>> = Mutex::new(HashMap::new());
        }
        let mut lock = CURVES.lock().unwrap();
        let curve = lock.entry(TypeId::of::<R>()).or_insert_with(|| {
            Box::new(MHCurve::<R>::from_bh(&[
                (  1.0,       0.0),
                (  2.0, 1000000.0),
                // Falling
                (  0.0,  900000.0),
            ]))
        }).downcast_ref::<MHCurve<R>>().unwrap();
        unsafe { std::mem::transmute::<&MHCurve<R>, &'static MHCurve<R>>(curve) }
    }
}

impl<R: Real> Material<R> for MinimalSquare<R> {
    fn step_b(&mut self, context: &CellState<R>, delta_b: Vec3<R>) {
        step_linear_ferro(&mut self.m, Self::curve(), context, delta_b)
    }
    fn m(&self) -> Vec3<R> {
        self.m
    }
    fn step_parameters_mut<'a>(&'a mut self) -> StepParametersMut<'a, R> {
        StepParametersMut::default().with_conductivity(Vec3::uniform(1e-3))
    }
}


#[cfg(test)]
mod test {
    use super::*;
    fn mh_curve_for_test() -> MHCurve<f32> {
        MHCurve::new(&[
            // rising
            Vec2::new( 10.0,    0.0),
            Vec2::new( 20.0,  100.0),
            Vec2::new( 30.0,  150.0),
            // falling
            Vec2::new(  0.0,  120.0),
            // negative rising
            Vec2::new(-10.0,    0.0),
            Vec2::new(-20.0, -100.0),
            Vec2::new(-30.0, -150.0),
            // negative falling
            Vec2::new(  0.0, -120.0),
        ])
    }

    fn assert_step_toward_symmetric(h: f32, m: f32, target_mh: f32, target: Result<Vec2<f32>, Vec2<f32>>) {
        let curve = mh_curve_for_test();
        let target = match target {
            Ok(v) => Ok(v),
            Err(v) => Err(v),
        };
        let neg_target = match target {
            Ok(v) => Ok(-v),
            Err(v) => Err(-v),
        };
        assert_eq!(curve.step_toward(h, m, target_mh), target);
        assert_eq!(curve.step_toward(-h, -m, -target_mh), neg_target);
    }

    fn assert_move_to_symmetric(h: f32, m: f32, target_mh: f32, target: (f32, f32)) {
        let curve = mh_curve_for_test();
        assert_eq!(curve.move_to(h, m, target_mh), target);
        assert_eq!(curve.move_to(-h, -m, -target_mh), (-target.0, -target.1));
    }

    #[test]
    fn mh_curve_move_from_inner_to_inner() {
        assert_step_toward_symmetric(0.0, 0.0, 5.0, Ok(Vec2::new(5.0, 0.0)));
        assert_step_toward_symmetric(0.0, 5.0, 10.0, Ok(Vec2::new(5.0, 5.0)));

        assert_step_toward_symmetric(-5.0, 5.0, -3.0, Ok(Vec2::new(-8.0, 5.0)));
        assert_step_toward_symmetric(-5.0, 5.0, 7.0, Ok(Vec2::new(2.0, 5.0)));

        assert_step_toward_symmetric(5.0, -5.0, -3.0, Ok(Vec2::new(2.0, -5.0)));
        assert_step_toward_symmetric(5.0, -5.0, 3.0, Ok(Vec2::new(8.0, -5.0)));
    }

    #[test]
    fn mh_curve_magnetize_along_edge() {
        // start of segment NOOP
        assert_step_toward_symmetric(10.0, 0.0, 10.0, Ok(Vec2::new(10.0, 0.0)));
        // start of segment to middle of segment
        assert_step_toward_symmetric(10.0, 0.0, 32.0, Ok(Vec2::new(12.0, 20.0)));
        // middle of segment NOOP
        assert_step_toward_symmetric(12.0, 20.0, 32.0, Ok(Vec2::new(12.0, 20.0)));
        // middle of segment to middle of segment
        assert_step_toward_symmetric(12.0, 20.0, 54.0, Ok(Vec2::new(14.0, 40.0)));
        // middle of segment to end of segment
        assert_step_toward_symmetric(12.0, 20.0, 120.0, Err(Vec2::new(20.0, 100.0)));
    }

    #[test]
    fn mh_curve_demagnetize_along_edge() {
        // start of segment NOOP
        assert_step_toward_symmetric(30.0, 150.0, 180.0, Ok(Vec2::new(30.0, 150.0)));
        // start of segment to middle of segment
        assert_step_toward_symmetric(30.0, 150.0, 160.0, Ok(Vec2::new(20.0, 140.0)));
        // middle of segment NOOP
        assert_step_toward_symmetric(20.0, 140.0, 160.0, Ok(Vec2::new(20.0, 140.0)));
        // middle of segment to middle of segment
        assert_step_toward_symmetric(20.0, 140.0, 140.0, Ok(Vec2::new(10.0, 130.0)));
        // middle of segment to end of segment
        assert_step_toward_symmetric(20.0, 140.0, 120.0, Err(Vec2::new(0.0, 120.0)));
    }

    #[test]
    fn mh_curve_magnetize_across_edges() {
        // Rising from start to middle
        assert_move_to_symmetric(10.0, 0.0, 132.0, (22.0, 110.0));
        // Rising from start to saturation
        assert_move_to_symmetric(10.0, 0.0, 180.0, (30.0, 150.0));
        // Rising from start to post-saturation
        assert_move_to_symmetric(10.0, 0.0, 400.0, (250.0, 150.0));
        // Rising from negative saturation to start
        assert_move_to_symmetric(-30.0, -150.0, 10.0, (10.0, 0.0));
        // Rising from negative post-saturation to start
        assert_move_to_symmetric(-250.0, -150.0, 10.0, (10.0, 0.0));
        // Rising from negative middle to middle
        assert_move_to_symmetric(-22.0, -110.0, 132.0, (22.0, 110.0));
    }

    #[test]
    fn mh_curve_demagnetize_across_edges() {
        // Falling from saturation to start
        assert_move_to_symmetric(30.0, 150.0, 120.0, (0.0, 120.0));
        // Falling from post-saturation to post-saturation
        assert_move_to_symmetric(250.0, 150.0, 200.0, (50.0, 150.0));
        // Falling from post-saturation to saturation
        assert_move_to_symmetric(250.0, 150.0, 180.0, (30.0, 150.0));
        // Falling from post-saturation to start
        assert_move_to_symmetric(250.0, 150.0, 120.0, (0.0, 120.0));
        // Falling from post-saturation to negative saturation
        assert_move_to_symmetric(250.0, 150.0, -180.0, (-30.0, -150.0));
        // Falling from post-saturation to negative post-saturation
        assert_move_to_symmetric(250.0, 150.0, -400.0, (-250.0, -150.0));
        // Falling from interior to middle
        assert_move_to_symmetric(28.0, 130.0, 140.0, (10.0, 130.0));
        // Falling from interior to middle
        assert_move_to_symmetric(28.0, 130.0, 130.0, (5.0, 125.0));
    }

    /// Float rounding would cause `inf`s, which manifested as infinite looping.
    #[test]
    fn regression_no_convergence_3r1() {
        let curve = Ferroxcube3R1::curve();
        curve.move_to(-202.04596, -278400.53, -278748.66);
    }
}
