use serde::{Deserialize, Serialize};

use crate::geom::Index;
use coremem_types::vec::{Vec3, Vec3u};

/// hide the actual spirv backend structures inside a submodule to make their use/boundary clear.
mod ffi {
    pub use spirv_backend::entry_points;
    pub use spirv_backend::sim::SerializedSimMeta;
    pub use spirv_backend::support::Optional;
}

// conversion traits for types defined cross-lib
pub trait IntoFfi {
    type Ffi;
    fn into_ffi(self) -> Self::Ffi;
}

pub trait IntoLib {
    type Lib;
    fn into_lib(self) -> Self::Lib;
}

pub trait Identity { }
impl<T: Identity> IntoFfi for T {
    type Ffi = Self;
    fn into_ffi(self) -> Self::Ffi {
        self
    }
}
impl<T: Identity> IntoLib for T {
    type Lib = Self;
    fn into_lib(self) -> Self::Lib {
        self
    }
}

impl Identity for f32 {}
impl<'a> Identity for &'a str {}
impl<T0: Identity, T1: Identity> Identity for (T0, T1) {}
impl Identity for Vec3u {}
impl<T: Identity> Identity for Vec3<T> {}

impl<L: IntoFfi> IntoFfi for Option<L>
    where L::Ffi: Default
{
    type Ffi = ffi::Optional<L::Ffi>;
    fn into_ffi(self) -> Self::Ffi {
        match self {
            Some(s) => ffi::Optional::some(s.into_ffi()),
            None => ffi::Optional::none(),
        }
    }
}

impl<F: Copy + IntoLib> IntoLib for ffi::Optional<F> {
    type Lib = Option<F::Lib>;
    fn into_lib(self) -> Self::Lib {
        if self.is_some() {
            Some(self.unwrap().into_lib())
        } else {
            None
        }
    }
}


// TODO: if we lift this type out of the spirv crate we can derive serde based on a feature flag
// this is bitwise- and type-compatible with the spirv SimMeta, except we need serde traits
#[derive(Clone, Default, Serialize, Deserialize)]
pub struct SimMeta<R> {
    pub(crate) dim: Index,
    pub(crate) inv_feature_size: R,
    pub(crate) time_step: R,
    pub(crate) feature_size: R,
}

impl<R: IntoFfi> IntoFfi for SimMeta<R> {
    type Ffi = ffi::SerializedSimMeta<R::Ffi>;
    fn into_ffi(self) -> Self::Ffi {
        Self::Ffi {
            dim: self.dim.0.into_ffi(),
            inv_feature_size: self.inv_feature_size.into_ffi(),
            time_step: self.time_step.into_ffi(),
            feature_size: self.feature_size.into_ffi(),
        }
    }
}

// FUNCTION BINDINGS
pub fn entry_points<L: 'static>() -> Option<(&'static str, &'static str)>
{
    ffi::entry_points::<L>().into_lib()
}
