use futures::FutureExt as _;
use ndarray::{self, Array3};
use serde::{Deserialize, Serialize};
use std::borrow::Cow;
use std::num::NonZeroU64;
use std::sync::Arc;
use wgpu;
use wgpu::util::DeviceExt as _;
use log::{info, trace, warn};

use crate::geom::{Coord, Index, Meters};
use crate::real::Real as _;
use crate::sim::{CellStateWithM, GenericSim, MaterialSim, Sample, SampleableSim};
use crate::stim::AbstractStimulus;
use crate::types::vec::Vec3;
use coremem_types::mat::{FullyGenericMaterial, Material};

mod bindings;
pub use bindings::{entry_points, IntoFfi as _, SimMeta};

/// Wrapper around an inner state object which offloads stepping onto a spirv backend (e.g. GPU).
#[derive(Clone, Default, Serialize, Deserialize)]
pub struct SpirvSim<M=FullyGenericMaterial<f32>>
{
    // TODO: make this generic over R
    meta: SimMeta<f32>,
    e: Vec<Vec3<f32>>,
    h: Vec<Vec3<f32>>,
    m: Vec<Vec3<f32>>,
    mat: Vec<M>,
    step_no: u64,
    // hidden behind an Arc to allow for cheap clones.
    // XXX not confident that wgpu is actually properly synchronized for us to omit a Mutex here
    // though.
    #[serde(skip)]
    wgpu: Option<Arc<WgpuData>>,
}

struct WgpuData {
    step_bind_group_layout: wgpu::BindGroupLayout,
    step_e_pipeline: wgpu::ComputePipeline,
    step_h_pipeline: wgpu::ComputePipeline,
    device: wgpu::Device,
    queue: wgpu::Queue,
}

impl WgpuData {
    pub fn new<M: 'static>(volume: u64) -> Self
    {
        info!("WgpuData::new({})", volume);
        use std::mem::size_of;
        let max_elem_size = size_of::<M>().max(size_of::<Vec3<f32>>());
        let max_buf_size = volume * max_elem_size as u64 + 0x1000;
        let entry_names = entry_points::<M>().unwrap_or(("invalid_mat", "invalid_mat"));
        let (device, queue) = futures::executor::block_on(open_device(max_buf_size));
        let shader_binary = get_shader();
        let shader_module = unsafe { device.create_shader_module_spirv(&shader_binary) };
        let (step_bind_group_layout, step_h_pipeline, step_e_pipeline) = make_pipelines(&device, &shader_module, entry_names);
        Self {
            step_bind_group_layout,
            step_h_pipeline,
            step_e_pipeline,
            device,
            queue,
        }
    }
}

impl Default for WgpuData {
    fn default() -> Self {
        Self::new::<FullyGenericMaterial<f32>>(0)
    }
}

impl<M> MaterialSim for SpirvSim<M>
where
// TODO: extraneous bounds?
    M: Send + Sync + Material<f32> + PartialEq + Clone + 'static
{
    type Material = M;

    fn put_material<C: Coord, M2: Into<Self::Material>>(&mut self, pos: C, mat: M2) {
        let pos_sim = pos.to_index(self.feature_size());
        let flat_idx = self.flat_index(pos_sim).unwrap();
        self.mat[flat_idx] = mat.into();
    }

    fn get_material<C: Coord>(&self, pos: C) -> Self::Material {
        let pos_sim = pos.to_index(self.feature_size());
        let flat_idx = self.flat_index(pos_sim).unwrap();
        self.mat[flat_idx].clone()
    }
}

impl<M> SampleableSim for SpirvSim<M>
where
    M: Send + Sync + Material<f32>
{
    fn sample(&self, pos: Meters) -> Sample {
        // TODO: smarter sampling than nearest neighbor?
        let pos_sim = pos.to_index(self.feature_size());
        let flat_idx = self.flat_index(pos_sim);
        match flat_idx {
            Some(idx) => Sample {
                state: CellStateWithM {
                    h: self.h[idx],
                    e: self.e[idx],
                    m: self.m[idx],
                },
                conductivity: self.mat[idx].conductivity(),
            },
            None => Default::default(),
        }
    }

    fn size(&self) -> Index {
        self.meta.dim
    }
    fn feature_size(&self) -> f32 {
        self.meta.feature_size
    }
    fn timestep(&self) -> f32 {
        self.meta.time_step
    }
    fn step_no(&self) -> u64 {
        self.step_no
    }
}

impl<M> GenericSim for SpirvSim<M>
where
    M: Send + Sync + Material<f32> + 'static
{
    fn step_multiple<S: AbstractStimulus>(&mut self, num_steps: u32, s: &S) {
        self.step_spirv(num_steps, s);
        self.step_no += num_steps as u64;
    }

    fn impulse_e_meters(&mut self, pos: Meters, amount: Vec3<f32>) {
        let idx = pos.to_index(self.feature_size());
        let flat_idx = self.flat_index(idx).unwrap();
        self.e[flat_idx] += amount;
    }

    fn impulse_h_meters(&mut self, pos: Meters, amount: Vec3<f32>) {
        let idx = pos.to_index(self.feature_size());
        let flat_idx = self.flat_index(idx).unwrap();
        self.h[flat_idx] += amount;
    }
    fn impulse_b_meters(&mut self, _pos: Meters, _amount: Vec3<f32>) {
        unimplemented!()
    }
}

impl<M> SpirvSim<M>
    where M: Default + 'static
{
    pub fn new(size: Index, feature_size: f32) -> Self {
        Self::new_with_wgpu_handle(size, feature_size, Some(Arc::new(WgpuData::new::<M>(size.volume()))))
    }

    pub fn new_no_wgpu(size: Index, feature_size: f32) -> Self {
        Self::new_with_wgpu_handle(size, feature_size, None)
    }

    fn new_with_wgpu_handle(size: Index, feature_size: f32, wgpu: Option<Arc<WgpuData>>) -> Self {
        info!("SpirvSim::new({:?}, {})", size, feature_size);
        let flat_size = size.volume() as usize;
        if flat_size * std::mem::size_of::<M>() >= 0x40000000 {
            warn!("simulation size ({:?}) is greater than what is tested for: anything could happen!", size);
        }
        if flat_size > spirv_backend::support::MAX_UNSIZED_ARRAY {
            warn!(
                "simulation size ({:?} elements) is too large (limit: {})!",
                flat_size,
                spirv_backend::support::MAX_UNSIZED_ARRAY
            );
        }
        let courant = 0.577;
        let time_step = feature_size / f32::c() * courant;
        let mut mat = Vec::new();
        mat.resize_with(flat_size, Default::default);
        Self {
            meta: SimMeta {
                dim: size,
                inv_feature_size: 1.0/feature_size,
                time_step,
                feature_size,
            },
            e: vec![Vec3::zero(); flat_size],
            h: vec![Vec3::zero(); flat_size],
            m: vec![Vec3::zero(); flat_size],
            mat,
            step_no: 0,
            wgpu,
        }
    }
}

impl<M> SpirvSim<M> {
    fn flat_index(&self, idx: Index) -> Option<usize> {
        let dim = self.meta.dim;
        if idx.x() < dim.x() && idx.y() < dim.y() && idx.z() < dim.z() {
            Some(((idx.z()*dim.y() + idx.y())*dim.x() + idx.x()) as usize)
        } else {
            None
        }
    }
}

impl<M> SpirvSim<M>
where
    M: Send + Sync + Material<f32> + 'static
{
    #[allow(unused)] // used for test
    fn apply_stimulus(&mut self, stim: &dyn AbstractStimulus) {
        trace!("apply_stimulus begin");
        iterate_stim(stim, self.size(), self.feature_size(), self.time(), self.timestep(), |pos_idx, value_e, value_h| {
            let flat_idx = self.flat_index(pos_idx).unwrap();
            self.e[flat_idx] += value_e;
            self.h[flat_idx] += value_h;
        });
        trace!("apply_stimulus end");
    }

    fn eval_stimulus<S: AbstractStimulus>(&self, stim: &S)
        -> (Array3<Vec3<f32>>, Array3<Vec3<f32>>)
    {
        trace!("eval_stimulus begin");
        let dim = self.size();
        let feature_size = self.feature_size();
        let t_sec = self.time();
        let timestep = self.timestep();

        // TODO(perf): do this in one loop!
        let e = ndarray::Zip::from(ndarray::indices(
            [dim.z() as usize, dim.y() as usize, dim.x() as usize]
        )).par_map_collect(|(z, y, x)| {
            let pos_idx = Index::new(x as _, y as _, z as _);
            let pos_meters = pos_idx.to_meters(feature_size);
            let (density_e, _density_h) = stim.at(t_sec, pos_meters);
            density_e * timestep
        });
        let h = ndarray::Zip::from(ndarray::indices(
            [dim.z() as usize, dim.y() as usize, dim.x() as usize]
        )).par_map_collect(|(z, y, x)| {
            let pos_idx = Index::new(x as _, y as _, z as _);
            let pos_meters = pos_idx.to_meters(feature_size);
            let (_density_e, density_h) = stim.at(t_sec, pos_meters);
            density_h * timestep
        });
        trace!("eval_stimulus end");
        (e, h)
    }

    fn step_spirv<S: AbstractStimulus>(&mut self, num_steps: u32, stim: &S) {
        let size = self.size();
        let (stim_cpu_e, stim_cpu_h) = self.eval_stimulus(stim);
        let field_bytes = size.product_sum() * 12;

        let sim_meta = [self.meta.clone().into_ffi()];

        self.wgpu.get_or_insert_with(
            || Arc::new(WgpuData::new::<M>(size.volume()))
         );
        let wgpu = &**self.wgpu.as_ref().unwrap();
        let device = &wgpu.device;
        let queue = &wgpu.queue;
        let step_bind_group_layout = &wgpu.step_bind_group_layout;
        let step_e_pipeline = &wgpu.step_e_pipeline;
        let step_h_pipeline = &wgpu.step_h_pipeline;
        let sim_meta_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side simulation metadata"),
            contents: to_bytes(&sim_meta[..]),
            usage: wgpu::BufferUsages::STORAGE,
        });

        let stim_e_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side stimulus e field"),
            contents: to_bytes(stim_cpu_e.as_slice().unwrap()),
            usage: wgpu::BufferUsages::STORAGE
        });

        let stim_h_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side stimulus h field"),
            contents: to_bytes(stim_cpu_h.as_slice().unwrap()),
            usage: wgpu::BufferUsages::STORAGE
        });

        let mat_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side materials matrix"),
            contents: to_bytes(self.mat.as_slice()),
            // Can be used by the GPU and copied back to the CPU
            usage: wgpu::BufferUsages::STORAGE,
        });

        let e_field_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side in/out e field"),
            contents: to_bytes(self.e.as_slice()),
            usage: wgpu::BufferUsages::STORAGE.union(wgpu::BufferUsages::COPY_SRC),
        });

        let h_field_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side in/out h field"),
            contents: to_bytes(self.h.as_slice()),
            // Can be used by the GPU and copied back to the CPU
            usage: wgpu::BufferUsages::STORAGE.union(wgpu::BufferUsages::COPY_SRC),
        });

        let m_field_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side in/out m field"),
            contents: to_bytes(self.m.as_slice()),
            // Can be used by the GPU and copied back to the CPU
            usage: wgpu::BufferUsages::STORAGE.union(wgpu::BufferUsages::COPY_SRC),
        });

        let e_readback_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("cpu-side copy of e output buffer"),
            size: field_bytes as wgpu::BufferAddress,
            // Can be read to the CPU, and can be copied from the shader's storage buffer
            usage: wgpu::BufferUsages::MAP_READ | wgpu::BufferUsages::COPY_DST,
            mapped_at_creation: false,
        });

        let h_readback_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("cpu-side copy of h output buffer"),
            size: field_bytes as wgpu::BufferAddress,
            // Can be read to the CPU, and can be copied from the shader's storage buffer
            usage: wgpu::BufferUsages::MAP_READ | wgpu::BufferUsages::COPY_DST,
            mapped_at_creation: false,
        });

        let m_readback_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("cpu-side copy of m output buffer"),
            size: field_bytes as wgpu::BufferAddress,
            // Can be read to the CPU, and can be copied from the shader's storage buffer
            usage: wgpu::BufferUsages::MAP_READ | wgpu::BufferUsages::COPY_DST,
            mapped_at_creation: false,
        });

        let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
            label: None,
            layout: &step_bind_group_layout,
            entries: &[
                wgpu::BindGroupEntry {
                    binding: 0,
                    resource: sim_meta_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 1,
                    resource: stim_e_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 2,
                    resource: stim_h_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 3,
                    resource: mat_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 4,
                    resource: e_field_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 5,
                    resource: h_field_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 6,
                    resource: m_field_buffer.as_entire_binding(),
                },
            ],
        });

        let workgroups = ((self.width()+3) / 4, (self.height()+3) / 4, (self.depth()+3) / 4);

        let mut encoder =
            device.create_command_encoder(&wgpu::CommandEncoderDescriptor { label: None });

        for _ in 0..num_steps {
            {
                let mut cpass = encoder.begin_compute_pass(&wgpu::ComputePassDescriptor { label: None });
                cpass.set_bind_group(0, &bind_group, &[]);
                cpass.set_pipeline(&step_e_pipeline);
                cpass.dispatch(workgroups.0, workgroups.1, workgroups.2);
            }

            {
                let mut cpass = encoder.begin_compute_pass(&wgpu::ComputePassDescriptor { label: None });
                cpass.set_bind_group(0, &bind_group, &[]);
                cpass.set_pipeline(&step_h_pipeline);
                cpass.dispatch(workgroups.0, workgroups.1, workgroups.2);
            }
        }

        encoder.copy_buffer_to_buffer(
            &e_field_buffer,
            0,
            &e_readback_buffer,
            0,
            field_bytes as _,
        );

        encoder.copy_buffer_to_buffer(
            &h_field_buffer,
            0,
            &h_readback_buffer,
            0,
            field_bytes as _,
        );

        encoder.copy_buffer_to_buffer(
            &m_field_buffer,
            0,
            &m_readback_buffer,
            0,
            field_bytes as _,
        );

        queue.submit(Some(encoder.finish()));

        let sim_e = &mut self.e;
        let e_readback_slice = e_readback_buffer.slice(..);
        let e_readback_future = e_readback_slice.map_async(wgpu::MapMode::Read).then(|_| async {
                sim_e.copy_from_slice(unsafe {
                    from_bytes(e_readback_slice.get_mapped_range().as_ref())
                });
                e_readback_buffer.unmap();
            });

        let sim_h = &mut self.h;
        let h_readback_slice = h_readback_buffer.slice(..);
        let h_readback_future = h_readback_slice.map_async(wgpu::MapMode::Read).then(|_| async {
            sim_h.copy_from_slice(unsafe {
                from_bytes(h_readback_slice.get_mapped_range().as_ref())
            });
            h_readback_buffer.unmap();
        });

        let sim_m = &mut self.m;
        let m_readback_slice = m_readback_buffer.slice(..);
        let m_readback_future = m_readback_slice.map_async(wgpu::MapMode::Read).then(|_| async {
            sim_m.copy_from_slice(unsafe {
                from_bytes(m_readback_slice.get_mapped_range().as_ref())
            });
            m_readback_buffer.unmap();
        });

        device.poll(wgpu::Maintain::Wait);

        futures::executor::block_on(futures::future::join(
            e_readback_future, futures::future::join(
                h_readback_future, m_readback_future)));
    }
}

fn iterate_stim<S: AbstractStimulus + ?Sized, F: FnMut(Index, Vec3<f32>, Vec3<f32>)>(
    stim: &S, dim: Index, feature_size: f32, t_sec: f32, timestep: f32, mut f: F
) {
    // TODO: parallelize
    for z in 0..dim.z() {
        for y in 0..dim.y() {
            for x in 0..dim.x() {
                let pos_idx = Index::new(x, y, z);
                let pos_meters = pos_idx.to_meters(feature_size);
                let (density_e, density_h) = stim.at(t_sec, pos_meters);
                let (value_e, value_h) = (density_e * timestep, density_h * timestep);
                f(pos_idx, value_e, value_h)
            }
        }
    }
}

/// Convert an arbitrary slice into a byte slice
fn to_bytes<T>(slice: &[T]) -> &[u8] {
    unsafe {
        std::slice::from_raw_parts(slice.as_ptr() as *const u8, slice.len() * std::mem::size_of::<T>())
    }
}

/// Convert a byte slice into a T slice
unsafe fn from_bytes<T>(slice: &[u8]) -> &[T] {
    let elem_size = std::mem::size_of::<T>();
    let new_len = slice.len() / elem_size;
    assert_eq!(new_len * elem_size, slice.len());
    std::slice::from_raw_parts(slice.as_ptr() as *const T, new_len)
}

/// Loads the shader
fn get_shader() -> wgpu::ShaderModuleDescriptorSpirV<'static> {
    let data = spirv_backend_runner::spirv_module();
    let spirv = Cow::Owned(wgpu::util::make_spirv_raw(&data).into_owned());

    wgpu::ShaderModuleDescriptorSpirV {
        label: None,
        source: spirv,
    }
}

async fn open_device(max_buf_size: u64) -> (wgpu::Device, wgpu::Queue) {
    // based on rust-gpu/examples/runners/wgpu/src/compute.rs:start_internal
    let instance = wgpu::Instance::new(wgpu::Backends::PRIMARY);
    info!("open_device: got instance");
    let adapter = instance
        .request_adapter(&wgpu::RequestAdapterOptions {
            power_preference: wgpu::PowerPreference::HighPerformance,
            force_fallback_adapter: false,
            compatible_surface: None,
        })
        .await
        .expect("Failed to find an appropriate adapter");
    info!("open_device: got adapter");

    // XXX not all adapters will support non-default limits, and it could
    // cause perf degradations even on the ones that do. May want to consider
    // folding some buffers together to avoid this.
    let mut limits = wgpu::Limits::default();
    //limits.max_bind_groups = 5;
    //limits.max_dynamic_storage_buffers_per_pipeline_layout = 5;
    limits.max_storage_buffers_per_shader_stage = 7;
    //limits.max_storage_buffer_binding_size = 128 MiB.
    //limits.max_storage_buffer_binding_size = 1024 * (1 << 20);
    limits.max_storage_buffer_binding_size = max_buf_size.try_into().unwrap();

    let (device, queue) = adapter
        .request_device(
            &wgpu::DeviceDescriptor {
                label: None,
                features: wgpu::Features::SPIRV_SHADER_PASSTHROUGH,
                limits,
            },
            None,
        )
        .await
        .expect("Failed to create device");
    info!("open_device: got device");
    (device, queue)
}

fn make_pipelines(device: &wgpu::Device, shader_module: &wgpu::ShaderModule, entry_names: (&'static str, &'static str)) -> (
    wgpu::BindGroupLayout, wgpu::ComputePipeline, wgpu::ComputePipeline
) {
    let bind_group_layout = device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
        label: None,
        entries: &[
            wgpu::BindGroupLayoutEntry {
                // meta
                binding: 0,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: true },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // stimulus(e)
                binding: 1,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: true },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // stimulus(h)
                binding: 2,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: true },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // materials
                binding: 3,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: true },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // e field
                binding: 4,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: false },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // h field
                binding: 5,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: false },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // m field
                binding: 6,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: false },
                },
            },
        ],
    });

    let pipeline_layout = device.create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
        label: None,
        bind_group_layouts: &[&bind_group_layout],
        push_constant_ranges: &[],
    });

    let compute_step_h_pipeline = device.create_compute_pipeline(&wgpu::ComputePipelineDescriptor {
        label: None,
        layout: Some(&pipeline_layout),
        module: shader_module,
        entry_point: entry_names.0,
    });

    let compute_step_e_pipeline = device.create_compute_pipeline(&wgpu::ComputePipelineDescriptor {
        label: None,
        layout: Some(&pipeline_layout),
        module: shader_module,
        entry_point: entry_names.1,
    });

    (bind_group_layout, compute_step_h_pipeline, compute_step_e_pipeline)
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::real::ToFloat as _;
    use crate::sim::SampleableSim;
    use crate::sim::legacy::{self, SimState};
    use crate::mat::{self, AnisomorphicConductor};
    fn mean_magnitude_e(sim: &dyn SampleableSim) -> f32 {
        (sim.map_sum_enumerated(|_pos: Index, cell| {
            cell.e().mag().to_f64()
        })/((sim.width() * sim.height() * sim.depth()) as f64)).to_f32()
    }
    fn mean_magnitude_h(sim: &dyn SampleableSim) -> f32 {
        (sim.map_sum_enumerated(|_pos: Index, cell| {
            cell.h().mag().to_f64()
        })/((sim.width() * sim.height() * sim.depth()) as f64)).to_f32()
    }
    fn mean_magnitude_m(sim: &dyn SampleableSim) -> f32 {
        (sim.map_sum_enumerated(|_pos: Index, cell| {
            cell.m().mag().to_f64()
        })/((sim.width() * sim.height() * sim.depth()) as f64)).to_f32()
    }

    fn assert_simstate_eq(real_state: &SpirvSim, ref_state: &SimState) {
        let rel_threshold = 1e-2;
        let mean_e = mean_magnitude_e(ref_state);
        let mean_h = mean_magnitude_h(ref_state);
        let mean_m = mean_magnitude_m(ref_state);
        let dim = real_state.size();

        for z in 0..dim.z() {
            for y in 0..dim.y() {
                for x in 0..dim.x() {
                    let idx = Index::new(x, y, z);
                    let real_cell = (real_state as &dyn SampleableSim).get(idx);
                    let ref_cell = (ref_state as &dyn SampleableSim).get(idx);

                    let diff_e = real_cell.e() - ref_cell.e();
                    let diff_h = real_cell.h() - ref_cell.h();
                    let diff_m = real_cell.m() - ref_cell.m();

                    // if x == 1 && y == 1 && z == 0 {
                    // if ref_cell.m() != Vec3::zero() {
                    //     println!("e {} v.s. {}", real_cell.e(), ref_cell.e());
                    //     println!("h {} v.s. {}", real_cell.h(), ref_cell.h());
                    //     println!("m {} v.s. {}", real_cell.m(), ref_cell.m());
                    //     println!("diff_e: {}; diff_h: {}, diff_m: {}", diff_e, diff_h, diff_m);
                    // }
                    assert!(
                        diff_e.mag() <= mean_e * rel_threshold,
                        "idx.e: {:?}. {:?} v.s. {:?}", idx, real_cell.e(), ref_cell.e()
                    );

                    assert!(
                        diff_h.mag() <= mean_h * rel_threshold,
                        "idx.h: {:?}. {:?} v.s. {:?}", idx, real_cell.h(), ref_cell.h()
                    );

                    assert!(
                        diff_m.mag() <= mean_m * rel_threshold || diff_m.mag() <= ref_cell.m().mag() * rel_threshold,
                        "idx.m: {:?}. {:?} v.s. {:?}", idx, real_cell.m(), ref_cell.m()
                    );
                }
            }
        }
    }

    mod cpu {
        //! Validate the spirv_backend implementation against the reference, but
        //! invoke it natively -- not through rgpu
        use super::*;
        use crate::sim::Meters;
        use crate::stim::{NoopStimulus, RngStimulus};
        use spirv_backend::{step_e_generic_material as step_e, step_h_generic_material as step_h, glam};

        fn do_step<S: AbstractStimulus>(state: &mut SpirvSim, stim: &S) {
            let meta = state.meta.clone();
            let meta_for_step = unsafe { std::mem::transmute(&meta) };
            let (stim_data_e, stim_data_h) = state.eval_stimulus(stim);
            let stim_e = unsafe { std::mem::transmute(&stim_data_e[[0, 0, 0]]) };
            let stim_h = unsafe { std::mem::transmute(&stim_data_h[[0, 0, 0]]) };
            let mat = unsafe { std::mem::transmute(&state.mat[0]) };
            let e: &mut _ = unsafe {
                std::mem::transmute(&mut state.e[0])
            };
            let h: &mut _ = unsafe {
                std::mem::transmute(&mut state.h[0])
            };
            let m: &mut _ = unsafe {
                std::mem::transmute(&mut state.m[0])
            };

            for z in 0..meta.dim.z() {
                for y in 0..meta.dim.y() {
                    for x in 0..meta.dim.x() {
                        step_e(glam::UVec3::new(x, y, z), meta_for_step, stim_e, stim_h, mat, e, h, m);
                    }
                }
            }

            for z in 0..meta.dim.z() {
                for y in 0..meta.dim.y() {
                    for x in 0..meta.dim.x() {
                        step_h(glam::UVec3::new(x, y, z), meta_for_step, stim_e, stim_h, mat, e, h, m);
                    }
                }
            }
        }

        #[test]
        fn smoke() {
            let mut ref_state = SimState::new(Index::new(8, 8, 8), 1e-3);
            let mut dut_state = SpirvSim::new_no_wgpu(Index::new(8, 8, 8), 1e-3);

            ref_state.step();
            do_step(&mut dut_state, &NoopStimulus);

            assert_simstate_eq(&dut_state, &ref_state);
        }

        #[test]
        fn smoke_large() {
            let mut state = SpirvSim::new_no_wgpu(Index::new(300, 304, 308), 1e-3);
            do_step(&mut state, &NoopStimulus);
        }

        #[test]
        fn smoke_not_multiple_of_4() {
            let mut state = SpirvSim::new_no_wgpu(Index::new(3, 2, 5), 1e-3);
            do_step(&mut state, &NoopStimulus);
        }

        fn test_same_explicit(mut ref_state: SimState, mut dut_state: SpirvSim, steps: u64) {
            for step in 0..steps {
                println!("step {}", step);
                ref_state.step();
                do_step(&mut dut_state, &NoopStimulus);

                assert_simstate_eq(&dut_state, &ref_state);
            }
        }

        fn test_same(seed: u64, steps: u64, size: Index) {
            let mut ref_state = SimState::new(size, 1e-3);
            ref_state.apply_stimulus(&RngStimulus::new(seed));

            let mut dut_state = SpirvSim::new_no_wgpu(size, 1e-3);
            dut_state.apply_stimulus(&RngStimulus::new(seed));

            test_same_explicit(ref_state, dut_state, steps);
        }

        #[test]
        fn same_1_step_unit() {
            test_same(0x1234, 1, Index::new(1, 1, 1));
        }

        #[test]
        fn same_1_step_tiny_x() {
            test_same(0x1234, 1, Index::new(2, 1, 1));
        }

        #[test]
        fn same_1_step_tiny_y() {
            let mut ref_state = SimState::new(Index::new(1, 2, 1), 1e-3);
            ref_state.impulse_e_meters(Meters::new(0.0, 0.0, 0.0), (1.0, 2.0, 3.0).into());
            let mut dut_state = SpirvSim::new_no_wgpu(Index::new(1, 2, 1), 1e-3);
            dut_state.impulse_e_meters(Meters::new(0.0, 0.0, 0.0), (1.0, 2.0, 3.0).into());
            test_same_explicit(ref_state, dut_state, 1);
        }

        #[test]
        fn same_1_step_tiny_z() {
            test_same(0x1234, 1, Index::new(1, 1, 2));
        }

        #[test]
        fn same_1_step_tiny_xyz() {
            test_same(0x1234, 1, Index::new(2, 2, 2));
        }

        #[test]
        fn same_1_step_tiny_xyz3() {
            test_same(0x1234, 1, Index::new(3, 3, 3));
        }

        #[test]
        fn same_1_step() {
            test_same(0x1234, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn same_1000_steps() {
            test_same(0x1234, 1000, Index::new(4, 4, 4));
        }

        #[test]
        fn same_1_step_larger() {
            test_same(0x1234, 1, Index::new(24, 20, 44));
        }

        #[test]
        fn same_100_steps_larger() {
            test_same(0x1234, 100, Index::new(24, 20, 44));
        }

        fn test_same_conductor(seed: u64, steps: u64, size: Index) {
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut ref_state = SimState::new(size, 1e-3);
            let mut dut_state = SpirvSim::new_no_wgpu(size, 1e-3);

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        let cond = Vec3::new(
                            rng.gen_range(0.0..1e6),
                            rng.gen_range(0.0..1e6),
                            rng.gen_range(0.0..1e6),
                        );
                        let m = AnisomorphicConductor::new(cond.cast());
                        ref_state.put_material(Index::new(x, y, z), m);
                        let m = AnisomorphicConductor::new(cond.cast());
                        dut_state.put_material(Index::new(x, y, z), m);
                    }
                }
            }

            ref_state.apply_stimulus(&RngStimulus::new(seed));
            dut_state.apply_stimulus(&RngStimulus::new(seed));
            test_same_explicit(ref_state, dut_state, steps);
        }

        #[test]
        fn conductor_1_step() {
            test_same_conductor(0x1234, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn conductor_100_steps_larger() {
            test_same_conductor(0x1234, 100, Index::new(96, 16, 8));
        }

        fn test_same_mb_ferromagnet(seed: u64, steps: u64, size: Index) {
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut ref_state = SimState::new(size, 1e-3);
            let mut dut_state = SpirvSim::new_no_wgpu(size, 1e-3);

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        let b_start = rng.gen_range(-1e-4..1e-4);
                        // if b_start is negative, we need the endpoint to be > |b_start|:
                        let b_range = (-2.0f32*b_start).max(0.0) + rng.gen_range(1e-5..1e-2);
                        // It would be weird for an increase in B to trigger a DECREASE in H.
                        // We have mu0(M+H) = B, so
                        // mu0(dM/dB = dH/dB) = 1
                        //   -> dH/dB = mu0^-1 - dM/dB > 0
                        let slope = rng.gen_range(0.0..0.9*f32::mu0_inv());
                        let b_end = b_start + b_range;
                        let m_max = b_range * slope;
                        ref_state.put_material(Index::new(x, y, z), legacy::mat::MBPgram::new(
                            b_start.cast(), b_end.cast(), m_max.cast()
                        ));
                        dut_state.put_material(Index::new(x, y, z), coremem_types::mat::MBPgram::new(
                            b_start, b_end, m_max
                        ));
                    }
                }
            }

            // XXX H stimulus on the ferromagnet behaves differently on the CPU than GPU
            // the CPU implementation is the more suspect one here (it doesn't consider material
            // parameters).
            ref_state.apply_stimulus(&RngStimulus::new_e(seed));
            dut_state.apply_stimulus(&RngStimulus::new_e(seed));

            test_same_explicit(ref_state, dut_state, steps);
        }

        #[test]
        fn mb_ferromagnet_1_step() {
            test_same_mb_ferromagnet(0x1234, 1, Index::new(4, 4, 4));
        }

        // XXX these tests probably failed because they were allowing negative mu_r values
        // #[test]
        // fn mb_ferromagnet_diff_repro() {
        //     let size = Index::new(3, 3, 3);
        //     let seed = 0;  // 0 => 28, 4 => 28
        //     let steps = 1000;
        //     let ref_mat = mat::MBPgram::new(0e-4, 9.02e-3, 1e4).into();
        //     let dut_mat = mat::MBPgram::new(0e-4, 9.02e-3, 1e4);
        //     let mut ref_state = SimState::new(size, 1e-3);
        //     let mut dut_state = SpirvSim::new(size, 1e-3);
        //     ref_state.put_material(Index::new(1, 1, 1), ref_mat);
        //     dut_state.put_material(Index::new(1, 1, 1), dut_mat);

        //     ref_state.apply_stimulus(&RngStimulus::new(seed));
        //     dut_state.apply_stimulus(&RngStimulus::new(seed));

        //     test_same_explicit(ref_state, dut_state, steps);
        // }

        // #[test]
        // fn mb_ferromagnet_diff_minimal_repro() {
        //     let size = Index::new(3, 3, 1);
        //     let seed = 0;  // 0 => 28, 4 => 28
        //     let steps = 1000;
        //     let ref_mat = mat::MBPgram::new(0e-4, 9.02e-3, 1e4).into();
        //     let dut_mat = mat::MBPgram::new(0e-4, 9.02e-3, 1e4);
        //     let mut ref_state = SimState::new(size, 1e-3);
        //     let mut dut_state = SpirvSim::new(size, 1e-3);
        //     ref_state.put_material(Index::new(1, 1, 0), ref_mat);
        //     dut_state.put_material(Index::new(1, 1, 0), dut_mat);

        //     ref_state.apply_stimulus(&RngStimulus::new(seed));
        //     dut_state.apply_stimulus(&RngStimulus::new(seed));

        //     test_same_explicit(ref_state, dut_state, steps);
        // }

        #[test]
        fn mb_ferromagnet_100_steps_larger() {
            test_same_mb_ferromagnet(0x1234, 100, Index::new(68, 16, 16));
        }

        fn test_smoke_mh_ferromagnet(seed: u64, steps: u64, size: Index) {
            // XXX This doesn't do anything, except make sure we don't crash!
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut dut_state = SpirvSim::new_no_wgpu(size, 1e-3);

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        let h_start = rng.gen_range(1.0..1000.0);
                        let mu_r = rng.gen_range(1.1..100000.0);
                        let m_max = rng.gen_range(0.0..1_000_000.0);
                        dut_state.put_material(Index::new(x, y, z), mat::MHPgram::new(h_start, mu_r, m_max));
                    }
                }
            }

            dut_state.apply_stimulus(&RngStimulus::new(seed));
            for step in 0..steps {
                println!("step {}", step);
                do_step(&mut dut_state, &NoopStimulus);
            }
        }

        #[test]
        fn mh_ferromagnet_smoke_1_step() {
            test_smoke_mh_ferromagnet(0x1234, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn mh_ferromagnet_smoke_100_steps_larger() {
            test_smoke_mh_ferromagnet(0x1234, 100, Index::new(68, 16, 16));
        }
    }
    mod rgpu {
        use super::*;
        use crate::stim::{NoopStimulus, RngStimulus, UniformStimulus};

        #[test]
        fn smoke_small() {
            let mut state: SpirvSim = SpirvSim::new(Index::new(8, 8, 8), 1e-3);
            state.step();
        }

        #[test]
        fn smoke_med_7bit() {
            let mut state: SpirvSim = SpirvSim::new(Index::new(124, 124, 124), 1e-3);
            state.step();
        }

        #[test]
        fn smoke_med128() {
            let mut state: SpirvSim = SpirvSim::new(Index::new(128, 128, 128), 1e-3);
            state.step();
        }

        #[test]
        fn smoke_med_23bit() {
            let mut state: SpirvSim = SpirvSim::new(Index::new(127, 256, 256), 1e-3);
            state.step();
        }

        #[test]
        fn smoke_med_0x800000_indexing() {
            let mut state: SpirvSim = SpirvSim::new(Index::new(128, 256, 256), 1e-3);
            state.step();
        }

        #[test]
        fn smoke_med_0xff00000_address_space() {
            let mut state: SpirvSim = SpirvSim::new(Index::new(170, 256, 256), 1e-3);
            state.step();
        }

        // #[test]
        // fn smoke_med_0x10080000_address_space() {
        //     let mut state = SpirvSim::new(Index::new(171, 256, 256), 1e-3);
        //     state.step();
        // }

        // #[test]
        // fn smoke_med_0xf00000_indexing() {
        //     let mut state = SpirvSim::new(Index::new(240, 256, 256), 1e-3);
        //     state.step();
        // }

        // #[test]
        // fn smoke_med252() {
        //     let mut state = SpirvSim::new(Index::new(252, 252, 252), 1e-3);
        //     state.step();
        // }

        #[test]
        fn smoke_large() {
            let mut state: SpirvSim = SpirvSim::new(Index::new(326, 252, 160), 1e-3);
            state.step();
        }

        #[test]
        fn smoke_not_multiple_of_4() {
            let mut state: SpirvSim = SpirvSim::new(Index::new(3, 2, 5), 1e-3);
            state.step();
        }

        fn test_same_explicit(mut ref_state: SimState, mut dut_state: SpirvSim, step_iters: u64, steps_per_iter: u32) {
            for _ in 0..step_iters {
                ref_state.step_multiple(steps_per_iter, &NoopStimulus);
                dut_state.step_multiple(steps_per_iter, &NoopStimulus);
                assert_simstate_eq(&dut_state, &ref_state);
            }
        }

        fn test_same(seed: u64, step_iters: u64, steps_per_iter: u32, size: Index) {
            let mut cpu_state = SimState::new(size, 1e-3);
            cpu_state.apply_stimulus(&RngStimulus::new(seed));

            let mut spirv_state = SpirvSim::new(size, 1e-3);
            spirv_state.apply_stimulus(&RngStimulus::new(seed));

            test_same_explicit(cpu_state, spirv_state, step_iters, steps_per_iter);
        }

        #[test]
        fn same_1_step() {
            test_same(0x1234, 1, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn same_1000_step() {
            test_same(0x1234, 1000, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn same_not_multiple_of_4() {
            test_same(0x1234, 100, 1, Index::new(3, 2, 5));
        }

        #[test]
        fn same_100_steps_larger() {
            test_same(0x1234, 100, 1, Index::new(24, 20, 44));
        }

        #[test]
        fn same_100_steps_of_10() {
            test_same(0x1234, 100, 10, Index::new(24, 20, 44));
        }

        fn test_same_conductor(seed: u64, step_iters: u64, steps_per_iter: u32, size: Index) {
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut ref_state = SimState::new(size, 1e-3);
            let mut dut_state: SpirvSim = SpirvSim::new(size, 1e-3);

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        let cond = Vec3::new(
                            rng.gen_range(0.0..1e6),
                            rng.gen_range(0.0..1e6),
                            rng.gen_range(0.0..1e6),
                        );
                        ref_state.put_material(Index::new(x, y, z), AnisomorphicConductor::new(cond.cast()));
                        dut_state.put_material(Index::new(x, y, z), AnisomorphicConductor::new(cond.cast()));
                    }
                }
            }

            ref_state.apply_stimulus(&RngStimulus::new(seed));
            dut_state.apply_stimulus(&RngStimulus::new(seed));

            test_same_explicit(ref_state, dut_state, step_iters, steps_per_iter);
        }

        #[test]
        fn conductor_1_step() {
            test_same_conductor(0x1234, 1, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn conductor_many_steps_larger() {
            test_same_conductor(0x1234, 100, 10, Index::new(96, 16, 8));
        }

        fn test_same_mb_ferromagnet(seed: u64, step_iters: u64, steps_per_iter: u32, size: Index) {
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut ref_state = SimState::new(size, 1e-3);
            let mut dut_state = SpirvSim::new(size, 1e-3);

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        let b_start = rng.gen_range(-1e-4..1e-4);
                        // if b_start is negative, we need the endpoint to be > |b_start|:
                        let b_range = (-2.0f32*b_start).max(0.0) + rng.gen_range(1e-5..1e-2);
                        // It would be weird for an increase in B to trigger a DECREASE in H.
                        // We have mu0(M+H) = B, so
                        // mu0(dM/dB = dH/dB) = 1
                        //   -> dH/dB = mu0^-1 - dM/dB > 0
                        let slope = rng.gen_range(0.0..0.9*f32::mu0_inv());
                        let b_end = b_start + b_range;
                        let m_max = b_range * slope;
                        ref_state.put_material(Index::new(x, y, z), legacy::mat::MBPgram::new(
                            b_start.cast(), b_end.cast(), m_max.cast()
                        ));
                        dut_state.put_material(Index::new(x, y, z), coremem_types::mat::MBPgram::new(
                            b_start, b_end, m_max
                        ));
                    }
                }
            }

            // XXX H stimulus on the ferromagnet behaves differently on the CPU than GPU
            // the CPU implementation is the more suspect one here (it doesn't consider material
            // parameters).
            ref_state.apply_stimulus(&RngStimulus::new_e(seed));
            dut_state.apply_stimulus(&RngStimulus::new_e(seed));

            test_same_explicit(ref_state, dut_state, step_iters, steps_per_iter);
        }

        #[test]
        fn mb_ferromagnet_1_step() {
            test_same_mb_ferromagnet(0x1234, 1, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn mb_ferromagnet_100_steps_larger() {
            test_same_mb_ferromagnet(0x1234, 10, 10, Index::new(96, 16, 8));
        }

        fn test_smoke_mh_ferromagnet(seed: u64, steps: u64, size: Index) {
            // XXX This doesn't do anything, except make sure we don't crash!
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut dut_state: SpirvSim = SpirvSim::new(size, 1e-3);

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        if rng.gen_range(0.0..1.0) < 0.5 {
                            let h_start = rng.gen_range(1.0..1000.0);
                            let mu_r = rng.gen_range(1.1..100000.0);
                            let m_max = rng.gen_range(0.0..1_000_000.0);
                            dut_state.put_material(Index::new(x, y, z), mat::MHPgram::new(h_start, mu_r, m_max));
                        } else if rng.gen_range(0.0..1.0) < 0.5 {
                            dut_state.put_material(Index::new(x, y, z), mat::IsomorphicConductor::new(rng.gen_range(0.0..1e7)));
                        }
                    }
                }
            }

            dut_state.apply_stimulus(&RngStimulus::new(seed));
            for step in 0..steps {
                println!("step {}", step);
                dut_state.step();
            }
        }

        #[test]
        fn mh_ferromagnet_smoke_1_step() {
            test_smoke_mh_ferromagnet(0x1234, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn mh_ferromagnet_smoke_100_steps_larger() {
            test_smoke_mh_ferromagnet(0x1234, 100, Index::new(328, 252, 160));
        }

        #[test]
        fn step_multiple_with_stim() {
            let size = Index::new(4, 12, 8);
            let mut ref_state = SimState::new(size, 1e-3);
            let mut dut_state = SpirvSim::new(size, 1e-3);
            let stim = UniformStimulus::new_e(Vec3::new(1.0e15, 2.0e15, -3.0e15));
            for _ in 0..5 {
                ref_state.step_multiple(100, &stim);
                dut_state.step_multiple(100, &stim);
                assert_simstate_eq(&dut_state, &ref_state);
            }
        }
    }
}
