use log::trace;
use serde::{de::DeserializeOwned, Serialize};

pub struct NoSupplier;
pub struct DiskCache<K, V, S=NoSupplier> {
    path: String,
    entries: Vec<(K, V)>,
    supplier: S,
}

impl<K: DeserializeOwned, V: DeserializeOwned> DiskCache<K, V, NoSupplier> {
    pub fn new(path: &str) -> Self {
        Self::new_with_supplier(path, NoSupplier)
    }
}

impl<K: DeserializeOwned, V: DeserializeOwned, S> DiskCache<K, V, S> {
    pub fn new_with_supplier(path: &str, supplier: S) -> Self {
        let entries = Self::load_from_disk(path).unwrap_or_default();
        Self {
            path: path.into(),
            entries,
            supplier,
        }
    }
    fn load_from_disk(path: &str) -> Option<Vec<(K, V)>> {
        let reader = std::io::BufReader::new(std::fs::File::open(path).ok()?);
        bincode::deserialize_from(reader).ok()
    }
}

impl<K: PartialEq, V, S> DiskCache<K, V, S> {
    pub fn get(&self, k: &K) -> Option<&V> {
        self.entries.iter().find(|(comp_k, _v): &&(K, V)| comp_k == k).map(|(_k, v)| v)
    }
}


impl<K: Serialize, V: Serialize, S> DiskCache<K, V, S> {
    pub fn insert(&mut self, k: K, v: V) {
        self.entries.push((k, v));
        self.flush();
    }
    fn flush(&self) {
        let writer = std::io::BufWriter::new(std::fs::File::create(&self.path).unwrap());
        bincode::serialize_into(writer, &self.entries).unwrap();
    }
}

impl<K: PartialEq + Serialize, V: Serialize + Clone, S> DiskCache<K, V, S> {
    pub fn get_or_insert_with<F: FnOnce() -> V>(&mut self, k: K, f: F) -> V {
        if let Some(v) = self.get(&k) {
            return v.clone();
        }

        let v = f();
        self.insert(k, v.clone());
        v
    }
}

impl<K: PartialEq + Serialize, V: Serialize + Clone, S: FnMut(&K) -> V> DiskCache<K, V, S> {
    pub fn get_or_insert_from_supplier(&mut self, k: K) -> V {
        if let Some(v) = self.get(&k) {
            trace!("get_or_insert_from_supplier hit");
            return v.clone();
        }

        trace!("get_or_insert_from_supplier miss");
        let v = (self.supplier)(&k);
        self.insert(k, v.clone());
        v
    }
}
