// use spirv_std::RuntimeArray;
use crate::support::{
    Array3, Array3Mut, ArrayHandle, ArrayHandleMut, Optional, UnsizedArray
};
use coremem_types::mat::Material;
use coremem_types::real::Real;
use coremem_types::vec::{Vec3, Vec3u};

#[derive(Copy, Clone)]
pub struct SerializedSimMeta<R> {
    pub dim: Vec3u,
    pub inv_feature_size: R,
    pub time_step: R,
    pub feature_size: R,
}


/// Whatever data we received from the host in their call to step_h
pub struct SerializedStepH<'a, R, M> {
    meta: &'a SerializedSimMeta<R>,
    stimulus_h: &'a UnsizedArray<Vec3<R>>,
    material: &'a UnsizedArray<M>,
    e: &'a UnsizedArray<Vec3<R>>,
    h: &'a mut UnsizedArray<Vec3<R>>,
    m: &'a mut UnsizedArray<Vec3<R>>,
}

impl<'a, R, M> SerializedStepH<'a, R, M> {
    pub fn new(
        meta: &'a SerializedSimMeta<R>,
        stimulus_h: &'a UnsizedArray<Vec3<R>>,
        material: &'a UnsizedArray<M>,
        e: &'a UnsizedArray<Vec3<R>>,
        h: &'a mut UnsizedArray<Vec3<R>>,
        m: &'a mut UnsizedArray<Vec3<R>>,
    ) -> Self {
        Self {
            meta, stimulus_h, material, e, h, m
        }
    }
}

impl<'a, R: Real, M> SerializedStepH<'a, R, M> {
    pub fn index(self, idx: Vec3u) -> StepHContext<'a, R, M> {
        let dim = self.meta.dim;
        let stim_h_matrix = Array3::new(self.stimulus_h, dim);
        let mat_matrix = Array3::new(self.material, dim);
        let e = Array3::new(self.e, dim);
        let h = Array3Mut::new(self.h, dim);
        let m = Array3Mut::new(self.m, dim);

        let in_e = VolumeSamplePos {
            mid: e.get(idx).unwrap(),
            xp1: e.get(idx + Vec3u::unit_x()),
            yp1: e.get(idx + Vec3u::unit_y()),
            zp1: e.get(idx + Vec3u::unit_z()),
        };
        let out_h = h.into_mut_handle(idx);
        let out_m = m.into_mut_handle(idx);

        let mat = mat_matrix.into_handle(idx);

        StepHContext {
            inv_feature_size: self.meta.inv_feature_size,
            time_step: self.meta.time_step,
            stim_h: stim_h_matrix.get(idx).unwrap(),
            mat,
            in_e,
            out_h,
            out_m,
        }
    }
}

/// Whatever data we received from the host in their call to step_e
pub struct SerializedStepE<'a, R, M> {
    meta: &'a SerializedSimMeta<R>,
    stimulus_e: &'a UnsizedArray<Vec3<R>>,
    material: &'a UnsizedArray<M>,
    e: &'a mut UnsizedArray<Vec3<R>>,
    h: &'a UnsizedArray<Vec3<R>>,
}

impl<'a, R, M> SerializedStepE<'a, R, M> {
    pub fn new(
        meta: &'a SerializedSimMeta<R>,
        stimulus_e: &'a UnsizedArray<Vec3<R>>,
        material: &'a UnsizedArray<M>,
        e: &'a mut UnsizedArray<Vec3<R>>,
        h: &'a UnsizedArray<Vec3<R>>,
    ) -> Self {
        Self {
            meta, stimulus_e, material, e, h
        }
    }
}

impl<'a, R: Real, M> SerializedStepE<'a, R, M> {
    pub fn index(self, idx: Vec3u) -> StepEContext<'a, R, M> {
        let dim = self.meta.dim;
        let stim_e_matrix = Array3::new(self.stimulus_e, dim);
        let mat_matrix = Array3::new(self.material, dim);
        let e = Array3Mut::new(self.e, dim);
        let h = Array3::new(self.h, dim);

        let xm1 = if idx.x() == 0 {
            Optional::none()
        } else {
            h.get(idx - Vec3u::unit_x())
        };
        let ym1 = if idx.y() == 0 {
            Optional::none()
        } else {
            h.get(idx - Vec3u::unit_y())
        };
        let zm1 = if idx.z() == 0 {
            Optional::none()
        } else {
            h.get(idx - Vec3u::unit_z())
        };

        let in_h = VolumeSampleNeg {
            mid: h.get(idx).unwrap(),
            xm1,
            ym1,
            zm1,
        };
        let out_e = e.into_mut_handle(idx);

        let mat = mat_matrix.into_handle(idx);

        StepEContext {
            inv_feature_size: self.meta.inv_feature_size,
            time_step: self.meta.time_step,
            stim_e: stim_e_matrix.get(idx).unwrap(),
            mat,
            in_h,
            out_e,
        }
    }
}

/// Package the field vectors adjacent to some particular location.
/// Particular those at negative offsets from the midpoint.
/// This is used in step_e when looking at the H field deltas.
#[derive(Copy, Clone)]
struct VolumeSampleNeg<R> {
    mid: Vec3<R>,
    xm1: Optional<Vec3<R>>,
    ym1: Optional<Vec3<R>>,
    zm1: Optional<Vec3<R>>,
}

impl<R: Real> VolumeSampleNeg<R> {
    /// Calculate the delta in H values amongst this cell and its neighbors (left/up/out)
    fn delta_h(self) -> FieldDeltas<R> {
        let mid = self.mid;
        // let (dfy_dx, dfz_dx) = self.xm1.map(|xm1| {
        //     (mid.y() - xm1.y(), mid.z() - xm1.z())
        // }).unwrap_or_default();

        // let (dfx_dy, dfz_dy) = self.ym1.map(|ym1| {
        //     (mid.x() - ym1.x(), mid.z() - ym1.z())
        // }).unwrap_or_default();

        // let (dfx_dz, dfy_dz) = self.zm1.map(|zm1| {
        //     (mid.x() - zm1.x(), mid.y() - zm1.y())
        // }).unwrap_or_default();

        let (dfy_dx, dfz_dx) = if self.xm1.is_some() {
            (mid.y() - self.xm1.unwrap().y(), mid.z() - self.xm1.unwrap().z())
        } else {
            (R::zero(), R::zero())
        };

        let (dfx_dy, dfz_dy) = if self.ym1.is_some() {
            (mid.x() - self.ym1.unwrap().x(), mid.z() - self.ym1.unwrap().z())
        } else {
            (R::zero(), R::zero())
        };

        let (dfx_dz, dfy_dz) = if self.zm1.is_some() {
            (mid.x() - self.zm1.unwrap().x(), mid.y() - self.zm1.unwrap().y())
        } else {
            (R::zero(), R::zero())
        };
        
        FieldDeltas {
            dfy_dx,
            dfz_dx,
            dfx_dy,
            dfz_dy,
            dfx_dz,
            dfy_dz,
        }
    }
}


/// Package the field vectors adjacent to some particular location.
/// Particular those at positive offsets from the midpoint.
/// This is used in step_h when looking at the E field deltas.
#[derive(Copy, Clone)]
struct VolumeSamplePos<R> {
    mid: Vec3<R>,
    xp1: Optional<Vec3<R>>,
    yp1: Optional<Vec3<R>>,
    zp1: Optional<Vec3<R>>
}

impl<R: Real> VolumeSamplePos<R> {
    /// Calculate the delta in E values amongst this cell and its neighbors (right/down/in)
    fn delta_e(self) -> FieldDeltas<R> {
        let mid = self.mid;
        // let (dfy_dx, dfz_dx) = self.xp1.map(|xp1| {
        //     (xp1.y() - mid.y(), xp1.z() - mid.z())
        // }).unwrap_or_default();

        // let (dfx_dy, dfz_dy) = self.yp1.map(|yp1| {
        //     (yp1.x() - mid.x(), yp1.z() - mid.z())
        // }).unwrap_or_default();

        // let (dfx_dz, dfy_dz) = self.zp1.map(|zp1| {
        //     (zp1.x() - mid.x(), zp1.y() - mid.y())
        // }).unwrap_or_default();

        let (dfy_dx, dfz_dx) = if self.xp1.is_some() {
            (self.xp1.unwrap().y() - mid.y(), self.xp1.unwrap().z() - mid.z())
        } else {
            (R::zero(), R::zero())
        };

        let (dfx_dy, dfz_dy) = if self.yp1.is_some() {
            (self.yp1.unwrap().x() - mid.x(), self.yp1.unwrap().z() - mid.z())
        } else {
            (R::zero(), R::zero())
        };

        let (dfx_dz, dfy_dz) = if self.zp1.is_some() {
            (self.zp1.unwrap().x() - mid.x(), self.zp1.unwrap().y() - mid.y())
        } else {
            (R::zero(), R::zero())
        };
        
        FieldDeltas {
            dfy_dx,
            dfz_dx,
            dfx_dy,
            dfz_dy,
            dfx_dz,
            dfy_dz,
        }
    }
}

struct FieldDeltas<R> {
    dfy_dx: R,
    dfz_dx: R,
    dfx_dy: R,
    dfz_dy: R,
    dfx_dz: R,
    dfy_dz: R,
}

impl<R: Real> FieldDeltas<R> {
    fn nabla(self) -> Vec3<R> {
        Vec3::new(
            self.dfz_dy - self.dfy_dz,
            self.dfx_dz - self.dfz_dx,
            self.dfy_dx - self.dfx_dy,
        )
    }
}

pub struct StepEContext<'a, R, M> {
    inv_feature_size: R,
    time_step: R,
    stim_e: Vec3<R>,
    mat: ArrayHandle<'a, M>,
    /// Input field sampled near this location
    in_h: VolumeSampleNeg<R>,
    /// Handle to the output field at one specific index.
    out_e: ArrayHandleMut<'a, Vec3<R>>,
}

impl<'a, R: Real, M: Material<R>> StepEContext<'a, R, M> {
    pub fn step_e(mut self) {
        let twice_eps0 = R::twice_eps0();
        let deltas = self.in_h.delta_h();
        // \nabla x H
        let nabla_h = deltas.nabla() * self.inv_feature_size;
        // $\nabla x H = \epsilon_0 dE/dt + \sigma E$
        // no-conductivity version:
        //   let delta_e = nabla_h * (self.time_step * EPS0_INV);
        let sigma = self.mat.get_ref().conductivity();
        let e_prev = self.out_e.get();
        let delta_e = (nabla_h - e_prev.elem_mul(sigma)).elem_div(
            sigma*self.time_step + Vec3::uniform(twice_eps0)
        )*(R::two()*self.time_step);
        // println!("spirv-step_e delta_e: {:?}", delta_e);
        self.out_e.write(e_prev + delta_e + self.stim_e);
    }
}


pub struct StepHContext<'a, R, M> {
    inv_feature_size: R,
    time_step: R,
    stim_h: Vec3<R>,
    mat: ArrayHandle<'a, M>,
    /// Input field sampled near this location
    in_e: VolumeSamplePos<R>,
    /// Handle to the output field at one specific index.
    out_h: ArrayHandleMut<'a, Vec3<R>>,
    out_m: ArrayHandleMut<'a, Vec3<R>>,
}

impl<'a, R: Real, M: Material<R>> StepHContext<'a, R, M> {
    pub fn step_h(mut self) {
        let mu0 = R::mu0();
        let mu0_inv = R::mu0_inv();
        let deltas = self.in_e.delta_e();
        // println!("spirv-step_h delta_e_struct: {:?}", deltas);
        // \nabla x E
        let nabla_e = deltas.nabla() * self.inv_feature_size;
        // println!("spirv-step_h nabla_e: {:?}", nabla_e);
        let delta_b = nabla_e * (-self.time_step);

        // Relation between these is: B = mu0*(H + M)
        let old_h = self.out_h.get();
        let old_m = self.out_m.get();
        let old_b = (old_h + old_m) * mu0;

        let new_b = old_b + delta_b + self.stim_h * mu0;
        let mat = self.mat.get_ref();
        let new_m = mat.move_b_vec(old_m, new_b);
        let new_h = new_b * mu0_inv - new_m;
        // println!("spirv-step_h delta_h: {:?}", delta_h);
        self.out_h.write(new_h);
        self.out_m.write(new_m);
    }
}
