use crate::compound::peano::{P0, Peano, PNext};
use crate::compound::list::{self, Indexable, IntoList, List};

#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

/// implement for something which supports being called for this specific variant
pub trait VariantHandler<N: Peano, Arg, Output> {
    fn call(self, a: Arg) -> Output;
}

/// anything which can encode a discriminant up to *but not including* P
pub trait DiscriminantCodable<P: Peano>: Sized {
    fn decode_discr(&self) -> Discr<P>;
    fn encode_discr(d: Discr<P>) -> Self;
    fn set_discr(&mut self, d: Discr<P>) {
        *self = Self::encode_discr(d)
    }
}

/// discriminant which encodes up to *but not including* P.
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct Discr<P: Peano>(u32, P::Unit);

impl<P: Peano> DiscriminantCodable<P> for Discr<P> {
    fn decode_discr(&self) -> Self {
        *self
    }
    fn encode_discr(d: Discr<P>) -> Self {
        d
    }
}

impl<P: Peano> Discr<P> {
    pub fn new(u: u32) -> Self {
        assert!(u < P::VALUE);
        Self::new_unchecked(u)
    }
    pub fn value(&self) -> u32 {
        self.0
    }
    fn new_unchecked(u: u32) -> Self {
        Self(u, Default::default())
    }
}

pub trait DiscrDispatch<P: Peano> {
    fn dispatch<H: DiscrHandler<P, Output>, Output>(&self, h: H) -> Output;
}

impl<P: Peano> DiscrDispatch<PNext<P>> for Discr<PNext<P>>
where
    Discr<P>: DiscrDispatch<P>
{
    fn dispatch<H: DiscrHandler<PNext<P>, O>, O>(&self, h: H) -> O {
        match self.value() {
            // consider P=2: we want to dispatch values of 0 and 1, but handle 2:
            // so dispatch v < P
            v if v < P::VALUE => Discr::<P>::new_unchecked(v).dispatch(h.prev()),
            v if v == P::VALUE => h.call(),
            _ => unreachable!(),
        }
    }
}

impl DiscrDispatch<P0> for Discr<P0> {
    fn dispatch<H: DiscrHandler<P0, O>, O>(&self, _h: H) -> O {
        unreachable!()
    }
}

/// something which can be called with any value up to (but not including) N
pub trait DiscrHandler<N: Peano, R> {
    type PrevOrPanic: DiscrHandler<N::PrevOrZero, R>;
    /// called when the discriminant has value N-1
    fn call(self) -> R;
    /// discriminant is < N-1: dispatch to the next handler.
    /// in the case that N = 1, this path *should* be unreachable,
    /// so a panic would be allowed.
    fn prev(self) -> Self::PrevOrPanic;
}

/// helper used to call F with some (yet-to-be-determined) index of I
pub struct DispatchIndexable<I, F> {
    indexable: I,
    f: F,
}

impl<I, F> DispatchIndexable<I, F> {
    fn new(indexable: I, f: F) -> Self {
        Self { indexable, f }
    }
}

// base case: we tried to index all cases >= 0 and failed.
// this should be unreachable.
impl<'a, I, F, R> DiscrHandler<P0, R> for DispatchIndexable<I, F>
{
    type PrevOrPanic = Self;
    fn call(self) -> R {
        unreachable!()
    }
    fn prev(self) -> Self::PrevOrPanic {
        unreachable!()
    }
}

// inductive case: if we know how dispatch up through P, and the collection is P+1-indexable, then we can
// index up through P+1
impl<'a, I, F, P: Peano, R> DiscrHandler<PNext<P>, R> for DispatchIndexable<&'a I, F>
where
    I: Indexable<P>,
    I::Element: Copy,
    F: VariantHandler<P, I::Element, R>,
    Self: DiscrHandler<P, R>,
{
    type PrevOrPanic = Self;
    fn call(self) -> R {
        self.f.call(self.indexable.get())
    }
    fn prev(self) -> Self::PrevOrPanic {
        self
    }
}

// mutable indexing case: if we have a mutable handle to the Indexable,
// then assume the variants want to have mutable references to the items.
impl<'a, I, F, P: Peano, R> DiscrHandler<PNext<P>, R> for DispatchIndexable<&'a mut I, F>
where
    I: Indexable<P>,
    I::Element: 'a,
    F: VariantHandler<P, &'a mut I::Element, R>,
    Self: DiscrHandler<P, R>,
{
    type PrevOrPanic = Self;
    fn call(self) -> R {
        self.f.call(self.indexable.get_mut())
    }
    fn prev(self) -> Self::PrevOrPanic {
        self
    }
}

#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct Enum<D, L>(D, L);

/// Users should prefer this type rather than rely on the internal Enum struct implementation.
#[allow(dead_code)]
pub type InternallyDiscriminated<Args> = Enum<(), List<Args>>;

impl<P: Peano, L> Enum<(Discr<P>,), L> {
    #![allow(dead_code)]
    pub fn new<Variants>(v: Variants) -> Self
    where
        Variants: IntoList<List=L>,
        L: list::Meta<Length=P>,
    {
        Enum((Discr::default(),), v.into_list())
    }
}
impl<L> Enum<(), L> {
    #![allow(dead_code)]
    pub fn internally_discriminated<Variants>(v: Variants) -> Self
    where
        Variants: IntoList<List=L>,
    {
        Enum((), v.into_list())
    }
}

pub trait EnumRequirements {
    type NumVariants: Peano;
    fn decode_discr(&self) -> Discr<Self::NumVariants>;
    fn encode_discr(&mut self, d: Discr<Self::NumVariants>);
}

impl<D, L> EnumRequirements for Enum<(D,), L>
where
    D: DiscriminantCodable<<L as list::Meta>::Length>,
    L: list::Meta,
{
    type NumVariants = <L as list::Meta>::Length;
    fn decode_discr(&self) -> Discr<Self::NumVariants> {
        self.0.0.decode_discr()
    }
    fn encode_discr(&mut self, d: Discr<Self::NumVariants>) {
        self.0.0.set_discr(d)
    }
}

impl<L> EnumRequirements for Enum<(), L>
where
    L: list::Meta + Indexable<P0>,
    list::ElementAt<P0, L>: DiscriminantCodable<<L as list::Meta>::Length>,
{
    type NumVariants = <L as list::Meta>::Length;
    fn decode_discr(&self) -> Discr<Self::NumVariants> {
        self.1.get_ref().decode_discr()
    }
    fn encode_discr(&mut self, d: Discr<Self::NumVariants>) {
        self.1.get_mut().set_discr(d)
    }
}


impl<D, L> Enum<D, L>
where
    Self: EnumRequirements
{
    /// invoke the closure on the active variant, passing the variant by-value
    pub fn dispatch<'a, F, R>(&'a self, f: F) -> R
    where
        DispatchIndexable<&'a L, F>: DiscrHandler<<Self as EnumRequirements>::NumVariants, R>,
        // TODO: this trait bound shouldn't be necessary. Discr ALWAYS implements DiscrDispatch
        Discr<<Self as EnumRequirements>::NumVariants>: DiscrDispatch<<Self as EnumRequirements>::NumVariants>,
    {
        self.decode_discr().dispatch(DispatchIndexable::new(&self.1, f))
    }

    /// invoke the closure on the active variant, passing the variant by mutable reference
    #[allow(dead_code)]
    pub fn dispatch_mut<'a, F, R>(&'a mut self, f: F) -> R
    where
        DispatchIndexable<&'a mut L, F>: DiscrHandler<<Self as EnumRequirements>::NumVariants, R>,
        Discr<<Self as EnumRequirements>::NumVariants>: DiscrDispatch<<Self as EnumRequirements>::NumVariants>,
    {
        self.decode_discr().dispatch(DispatchIndexable::new(&mut self.1, f))
    }

    pub fn set<P>(&mut self, value: L::Element)
    where
        P: Peano,
        L: Indexable<P>,
    {
        self.encode_discr(Discr::new(P::VALUE));
        self.1.set(value);
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::compound::peano::{P1, P2, P3};
    use crate::compound::list::List;

    struct ReadReceiver;
    impl<P: Peano, T: TryInto<i32>> VariantHandler<P, T, i32> for ReadReceiver {
        fn call(self, v: T) -> i32 {
            unsafe {
                v.try_into().unwrap_unchecked()
            }
        }
    }

    struct AddIndexPlus5Receiver;
    impl<P: Peano, T: TryInto<i32>> VariantHandler<P, T, i32> for AddIndexPlus5Receiver {
        fn call(self, v: T) -> i32 {
            unsafe {
                v.try_into().unwrap_unchecked() + P::VALUE as i32 + 5
            }
        }
    }

    struct Add4Receiver;
    impl<P: Peano, T: TryInto<i32> + TryFrom<i32> + Copy> VariantHandler<P, &mut T, ()> for Add4Receiver {
        fn call(self, v: &mut T) -> () {
            unsafe {
                *v = ((*v).try_into().unwrap_unchecked() + 4).try_into().unwrap_unchecked();
            }
        }
    }

    #[test]
    fn dispatch() {
        let mut e: Enum<(Discr<P3>,), List<(u32, i32, u8)>> = Enum::default();
        assert_eq!(e.dispatch(AddIndexPlus5Receiver), 5);

        e.encode_discr(Discr::new(1));
        assert_eq!(e.dispatch(AddIndexPlus5Receiver), 6);

        e.encode_discr(Discr::new(2));
        assert_eq!(e.dispatch(AddIndexPlus5Receiver), 7);
    }

    #[test]
    fn dispatch_mut() {
        let mut e: Enum<(Discr<P3>,), List<(u32, i32, u8)>> = Enum::default();
        e.dispatch_mut(Add4Receiver);
        assert_eq!(e.dispatch(AddIndexPlus5Receiver), 9);

        e.encode_discr(Discr::new(1));
        assert_eq!(e.dispatch(AddIndexPlus5Receiver), 6);
    }

    #[test]
    fn set() {
        let mut e: Enum<(Discr<P3>,), List<(u32, i32, u8)>> = Enum::default();
        e.set::<P0>(2u32);
        assert_eq!(e.dispatch(ReadReceiver), 2);

        e.set::<P1>(3i32);
        assert_eq!(e.dispatch(ReadReceiver), 3);
    }

    #[derive(Copy, Clone, Default, PartialEq)]
    struct BoxedF32(f32);

    impl Into<i32> for BoxedF32 {
        fn into(self) -> i32 {
            self.0 as i32
        }
    }

    impl From<i32> for BoxedF32 {
        fn from(v: i32) -> Self {
            Self(v as f32)
        }
    }

    impl<P: Peano> DiscriminantCodable<P> for BoxedF32 {
        fn decode_discr(&self) -> Discr<P> {
            match self.0 {
                v if v < 0f32 => Discr::new((-v) as u32),
                _non_negative => Discr::new(0),
            }
        }
        fn encode_discr(d: Discr<P>) -> Self {
            Self(-(d.value() as f32))
        }
    }

    #[test]
    fn internal_discr() {
        type E = Enum<(), List<(BoxedF32, i32, u8)>>;
        assert_eq!(<E as EnumRequirements>::NumVariants::VALUE, 3);

        let mut e: E = Enum::default();
        assert_eq!(e.dispatch(ReadReceiver), 0);

        e.set::<P0>(BoxedF32(16f32));
        assert_eq!(e.dispatch(ReadReceiver), 16);

        e.set::<P1>(5);
        assert_eq!(e.dispatch(ReadReceiver), 5);

        e.set::<P2>(8);
        assert_eq!(e.dispatch(ReadReceiver), 8);

        e.set::<P0>(BoxedF32(0f32));
        assert_eq!(e.dispatch(ReadReceiver), 0);
    }

    #[test]
    fn new() {
        type E = Enum<(Discr<P2>,), List<(u32, i32)>>;
        assert_eq!(<E as EnumRequirements>::NumVariants::VALUE, 2);

        let e: E = Enum::new((5u32, 4i32));
        assert_eq!(e.dispatch(ReadReceiver), 5);

        let e = Enum::internally_discriminated((BoxedF32(4f32), -1i32));
        assert_eq!(e.dispatch(ReadReceiver), 4);
    }
}
