use crate::compound::list::{Indexable, Meta};
use crate::compound::peano::{P0, P1, P2, P3, Peano, PNext};

#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct Node<H, T> {
    head: H,
    tail: T,
}
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct Null;

impl<H, T> Node<H, T> {
    fn new(head: H, tail: T) -> Self {
        Self { head, tail }
    }

    pub fn get<P: Peano>(&self) -> <Self as Indexable<P>>::Element
    where
        Self: Indexable<P>,
        <Self as Indexable<P>>::Element: Copy,
    {
        Indexable::<P>::get(self)
    }
}

impl<E0, T> Indexable<P0> for Node<E0, T> {
    type Element = E0;
    fn get(&self) -> Self::Element where Self::Element: Copy {
        self.head
    }
    fn get_ref(&self) -> &Self::Element {
        &self.head
    }
    fn get_mut(&mut self) -> &mut Self::Element {
        &mut self.head
    }
    fn set(&mut self, v: Self::Element) {
        self.head = v;
    }
}

impl<E0, E1, T> Indexable<P1> for Node<E0, Node<E1, T>> {
    type Element = E1;
    fn get(&self) -> Self::Element where Self::Element: Copy {
        self.tail.head
    }
    fn get_ref(&self) -> &Self::Element {
        &self.tail.head
    }
    fn get_mut(&mut self) -> &mut Self::Element {
        &mut self.tail.head
    }
    fn set(&mut self, v: Self::Element) {
        self.tail.head = v;
    }
}

impl<E0, E1, E2, T> Indexable<P2> for Node<E0, Node<E1, Node<E2, T>>> {
    type Element = E2;
    fn get(&self) -> Self::Element where Self::Element: Copy {
        self.tail.tail.head
    }
    fn get_ref(&self) -> &Self::Element {
        &self.tail.tail.head
    }
    fn get_mut(&mut self) -> &mut Self::Element {
        &mut self.tail.tail.head
    }
    fn set(&mut self, v: Self::Element) {
        self.tail.tail.head = v;
    }
}

impl<E0, E1, E2, E3, T> Indexable<P3> for Node<E0, Node<E1, Node<E2, Node<E3, T>>>> {
    type Element = E3;
    fn get(&self) -> Self::Element where Self::Element: Copy {
        self.tail.tail.tail.head
    }
    fn get_ref(&self) -> &Self::Element {
        &self.tail.tail.tail.head
    }
    fn get_mut(&mut self) -> &mut Self::Element {
        &mut self.tail.tail.tail.head
    }
    fn set(&mut self, v: Self::Element) {
        self.tail.tail.tail.head = v;
    }
}

impl<E0> IntoList for (E0,) {
    type List = Node<E0, Null>;
    fn into_list(self) -> Self::List {
        Node::new(self.0, Null)
    }
}

impl<E0, E1> IntoList for (E0, E1) {
    type List = Node<E0, <(E1,) as IntoList>::List>;
    fn into_list(self) -> Self::List {
        Node::new(self.0, (self.1,).into_list())
    }
}

impl<E0, E1, E2> IntoList for (E0, E1, E2) {
    type List = Node<E0, <(E1, E2) as IntoList>::List>;
    fn into_list(self) -> Self::List {
        Node::new(self.0, (self.1, self.2).into_list())
    }
}

impl<E0, E1, E2, E3> IntoList for (E0, E1, E2, E3) {
    type List = Node<E0, <(E1, E2, E3) as IntoList>::List>;
    fn into_list(self) -> Self::List {
        Node::new(self.0, (self.1, self.2, self.3).into_list())
    }
}

pub trait IntoList {
    type List;
    fn into_list(self) -> Self::List;
}

impl<H> Meta for Node<H, Null> {
    type Length = P1;
}
impl<H, T: Meta> Meta for Node<H, T> {
    type Length = PNext<T::Length>;
}

/// these are exported for the convenience of potential consumers: not needed internally
pub(crate) mod exports {
    #![allow(dead_code)]
    use super::{IntoList, Node, Null};
    pub type List<Args> = <Args as IntoList>::List;
    pub type List1<E0> = Node<E0, Null>;
    pub type List2<E0, E1> = Node<E0, List1<E1>>;
    pub type List3<E0, E1, E2> = Node<E0, List2<E1, E2>>;
    pub type List4<E0, E1, E2, E3> = Node<E0, List3<E1, E2, E3>>;
}
