use crate::compound::peano::{P0, Peano, PeanoNonZero};

#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct Middle<H, T> {
    head: H,
    tail: T,
}

#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct Terminal<H> {
    head: H,
    // XXX: needed to handle ZSTs in spirv, else we can't hand out a reference to &self
    _pad: u32,
}

impl<H, T> Middle<H, T> {
    fn new(head: H, tail: T) -> Self {
        Middle { head, tail }
    }
}

impl<H> Terminal<H> {
    fn new(head: H) -> Self {
        Terminal { head, _pad: Default::default() }
    }
}

// Self is a Superlist of L
pub trait Superlist<Distance: Peano> {
    type Of: ListOps;
    fn as_sublist(&self) -> &Self::Of;
}

impl<L: ListOps> Superlist<P0> for L {
    type Of = L;
    fn as_sublist(&self) -> &Self::Of {
        self
    }
}
// if our tail T0 is a Superlist<P-1> of T1,
// then we are a Superlist<P> of T1.
impl<H0, T0: ListOps, T1: ListOps, P: PeanoNonZero> Superlist<P> for Middle<H0, T0>
where
    T0: Superlist<P::Prev, Of=T1>
{
    type Of = T1;
    fn as_sublist(&self) -> &T1 {
        self.tail.as_sublist()
    }
}

pub trait ListOps {
    type Element;
    fn read(&self) -> Self::Element where Self::Element: Copy;

    fn index<P: Peano>(&self) -> &<Self as Superlist<P>>::Of
        where Self: Superlist<P>
    {
        self.as_sublist()
    }
    fn get<P: Peano>(&self) -> <<Self as Superlist<P>>::Of as ListOps>::Element
    where 
        Self: Superlist<P>,
        <<Self as Superlist<P>>::Of as ListOps>::Element: Copy

    {
        self.index::<P>().read()
    }
}

impl<H> ListOps for Terminal<H> {
    type Element = H;
    fn read(&self) -> Self::Element where Self::Element: Copy {
        self.head
    }
}

impl<H, T> ListOps for Middle<H, T> {
    type Element = H;
    fn read(&self) -> Self::Element where Self::Element: Copy {
        self.head
    }
}

pub trait IntoList {
    type List;
    fn into_list(self) -> Self::List;
}

impl<E0> IntoList for (E0,) {
    type List = Terminal<E0>;
    fn into_list(self) -> Self::List {
        Terminal::new(self.0)
    }
}
impl<E0, E1> IntoList for (E0, E1) {
    type List = Middle<E0, <(E1,) as IntoList>::List>;
    fn into_list(self) -> Self::List {
        Middle::new(self.0, (self.1,).into_list())
    }
}
impl<E0, E1, E2> IntoList for (E0, E1, E2) {
    type List = Middle<E0, <(E1, E2) as IntoList>::List>;
    fn into_list(self) -> Self::List {
        Middle::new(self.0, (self.1, self.2).into_list())
    }
}

pub type List<Args> = <Args as IntoList>::List;

#[cfg(test)]
mod test {
    use super::*;
    use crate::compound::peano::{P0, P1, P2};
    #[test]
    fn list_index() {
        let l = (1u32, 3f32, 4u32).into_list();
        assert_eq!(l.read(), 1u32);
        assert_eq!(l.index::<P0>().read(), 1u32);
        assert_eq!(l.index::<P1>().read(), 3f32);
        assert_eq!(l.index::<P2>().read(), 4u32);
    }
}
