use crate::real::{Real, ToFloat};
pub use crate::vecu::Vec3u;

use core::convert::From;
use core::iter::Sum;
use core::ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Neg, Sub};

#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

#[cfg(feature = "fmt")]
use core::fmt;

#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, Eq, PartialEq)]
pub struct Vec2<R> {
    pub x: R,
    pub y: R,
}

impl<R: Real> Add for Vec2<R> {
    type Output = Self;
    fn add(self, other: Self) -> Self {
        let mut ret = self.clone();
        ret += other;
        ret
    }
}

impl<R: Real> AddAssign for Vec2<R> {
    fn add_assign(&mut self, other: Self) {
        self.x += other.x;
        self.y += other.y;
    }
}

impl<R: Real> Neg for Vec2<R> {
    type Output = Self;
    fn neg(self) -> Self {
        Vec2 {
            x: -self.x,
            y: -self.y,
        }
    }
}

impl<R: Real> Sub for Vec2<R> {
    type Output = Self;
    fn sub(self, other: Self) -> Self {
        self + (-other)
    }
}

impl<R: Real> Mul<R> for Vec2<R> {
    type Output = Self;
    fn mul(self, s: R) -> Self {
        Vec2 {
            x: self.x * s,
            y: self.y * s,
        }
    }
}

impl<R> Into<(R, R)> for Vec2<R> {
    fn into(self) -> (R, R) {
        (self.x, self.y)
    }
}

impl<R> Vec2<R> {
    pub fn new(x: R, y: R) -> Self {
        Self {
            x, y
        }
    }
}

impl<R: Real> Vec2<R> {
    pub fn zero() -> Self {
        Self::new(R::zero(), R::zero())
    }

    pub fn unit() -> Self {
        Self::new(R::one(), R::one())
    }

    pub fn x(&self) -> R {
        self.x.into()
    }

    pub fn y(&self) -> R {
        self.y.into()
    }

    pub fn to_f32(&self) -> Vec2<f32> {
        self.cast()
    }

    pub fn cast<R2: Real>(&self) -> Vec2<R2> {
        Vec2::new(
            R2::from_primitive(self.x),
            R2::from_primitive(self.y),
        )
    }


    pub fn round(&self) -> Self {
        Vec2 {
            x: self.x.round(),
            y: self.y.round(),
        }
    }

    pub fn distance_sq(&self, other: Self) -> R {
        (*self - other).mag_sq()
    }

    pub fn mag_sq(&self) -> R {
        self.x*self.x + self.y*self.y
    }

    pub fn mag(&self) -> R {
        self.mag_sq().sqrt()
    }

    pub fn with_mag(&self, new_mag: R) -> Self {
        if new_mag.is_zero() {
            // avoid div-by-zero if self.mag() == 0 and new_mag == 0
            Vec2::new(R::zero(), R::zero())
        } else {
            let scale = new_mag / self.mag();
            *self * scale
        }
    }

    // /// Returns the angle of this point, (-pi, pi]
    // pub fn arg(&self) -> R {
    //     // self.y.atan2(self.x)
    //     self.y.to_f64().atan2(self.x.to_f64()).cast()
    // }

    // TODO test
    pub fn rotate(&self, angle: R) -> Self {
        let (sin, cos) = angle.sin_cos();
        let map_1x_0y = Vec2::new(cos, sin);
        let map_0x_1y = Vec2::new(-sin, cos);
        Self::new(
            map_1x_0y.x() * self.x() + map_0x_1y.x() * self.y(),
            map_1x_0y.y() * self.x() + map_0x_1y.y() * self.y(),
        )
    }
}

#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq, PartialOrd)]
pub struct Vec3<R> {
    pub(crate) x: R,
    pub(crate) y: R,
    pub(crate) z: R,
}

impl<R> Vec3<R> {
    pub fn to_tuple(self) -> (R, R, R) {
        self.into()
    }
}

impl<R: Real> Vec3<R> {
    pub fn new(x: R, y: R, z: R) -> Self {
        Self {
            x, y, z
        }
    }
    pub fn new_x(x: R) -> Self {
        Self {
            x,
            y: R::zero(),
            z: R::zero(),
        }
    }
    pub fn new_y(y: R) -> Self {
        Self {
            x: R::zero(),
            y,
            z: R::zero(),
        }
    }
    pub fn new_z(z: R) -> Self {
        Self {
            x: R::zero(),
            y: R::zero(),
            z,
        }
    }
    pub fn uniform(u: R) -> Self {
        Self {
            x: u,
            y: u,
            z: u,
        }
    }

    pub fn cast<R2: Real>(&self) -> Vec3<R2> {
        Vec3::new(
            R2::from_primitive(self.x),
            R2::from_primitive(self.y),
            R2::from_primitive(self.z)
        )
    }

    pub fn unit() -> Self {
        Self::new(R::one(), R::one(), R::one())
    }
    pub fn unit_x() -> Self {
        Self::new(R::one(), R::zero(), R::zero())
    }
    pub fn unit_y() -> Self {
        Self::new(R::zero(), R::one(), R::zero())
    }
    pub fn unit_z() -> Self {
        Self::new(R::zero(), R::zero(), R::one())
    }
    pub fn zero() -> Self {
        Self::new(R::zero(), R::zero(), R::zero())
    }

    pub fn set_x<R2: ToFloat>(&mut self, x: R2) {
        self.x = R::from_primitive(x);
    }
    pub fn set_y<R2: ToFloat>(&mut self, y: R2) {
        self.y = R::from_primitive(y);
    }
    pub fn set_z<R2: ToFloat>(&mut self, z: R2) {
        self.z = R::from_primitive(z);
    }
    pub fn x(&self) -> R {
        self.x
    }
    pub fn y(&self) -> R {
        self.y
    }
    pub fn z(&self) -> R {
        self.z
    }
    pub fn xy(&self) -> Vec2<R> {
        Vec2::new(self.x(), self.y())
    }
    pub fn with_xy(&self, xy: Vec2<R>) -> Self {
        Self::new(xy.x(), xy.y(), self.z())
    }

    pub fn distance(&self, other: Self) -> R {
        (*self - other).mag()
    }
    pub fn distance_sq(&self, other: Self) -> R {
        (*self - other).mag_sq()
    }
    pub fn mag(&self) -> R {
        self.mag_sq().sqrt()
    }
    pub fn mag_sq(&self) -> R {
        let Vec3 { x, y, z } = *self;
        x*x + y*y + z*z
    }
    pub fn component_sum(&self) -> R {
        self.x + self.y + self.z
    }

    pub fn dot(&self, other: Self) -> R {
        self.elem_mul(other).component_sum()
    }

    pub fn cross(&self, other: Self) -> Self {
        // det(   i    j    k  |
        //    |  s.x  s.y  s.z |
        //    |  o.x  o.y  o.z )
        let x = self.y * other.z - other.y * self.z;
        let y = other.z * self.z - self.x * other.z;
        let z = self.x * other.y - other.x * self.y;
        Self { x, y, z }
    }

    /// Return some non-zero vector whose dot product with this one yields 0.
    pub fn arbitrary_orthogonal_vector(&self) -> Self {
        if self.x != R::zero() || self.y != R::zero() {
            // dot = x*y + y*-x = 0
            Self::new(self.y, -self.x, R::zero())
        } else {
            assert!(self.z != R::zero(), "z is zero");
            Self::new(self.z, R::zero(), R::zero())
        }
    }

    /// Perform element-wise multiplication with `other`.
    pub fn elem_mul(&self, other: Self) -> Self {
        Self {
            x: self.x * other.x,
            y: self.y * other.y,
            z: self.z * other.z,
        }
    }

    pub fn elem_pow(&self, pow: R) -> Self {
        Self::new(
            self.x().powf(pow),
            self.y().powf(pow),
            self.z().powf(pow),
        )
    }

    /// Perform element-wise division with `other`.
    pub fn elem_div(&self, other: Self) -> Self {
        Self {
            x: self.x / other.x,
            y: self.y / other.y,
            z: self.z / other.z,
        }
    }

    /// Return (x^-1, y^-1, z^-1)
    pub fn elem_inv(&self) -> Self {
        let one = R::one();
        Self::new(one / self.x(), one / self.y(), one / self.z())
    }

    pub fn exp(&self) -> Self {
        Self::new(self.x().exp(), self.y().exp(), self.z().exp())
    }

    pub fn with_mag(&self, new_mag: R) -> Self {
        if new_mag.is_zero() {
            // avoid div-by-zero if self.mag() == 0 and new_mag == 0
            Self::zero()
        } else {
            let scale = new_mag / self.mag();
            *self * scale
        }
    }

    /// Normalizes to 1.0 length, unless it's the zero vector, in which case length remains zero.
    pub fn norm(&self) -> Self {
        if *self == Self::zero() {
            *self
        } else {
            self.with_mag(R::one())
        }
    }
    pub fn round(&self) -> Self {
        Self::new(self.x().round(), self.y().round(), self.z().round())
    }
    pub fn ceil(&self) -> Self {
        Self::new(self.x().ceil(), self.y().ceil(), self.z().ceil())
    }
    pub fn floor(&self) -> Self {
        Self::new(self.x().floor(), self.y().floor(), self.z().floor())
    }
}

impl<R> Into<(R, R, R)> for Vec3<R> {
    fn into(self) -> (R, R, R) {
        (self.x, self.y, self.z)
    }
}

impl<R> From<(R, R, R)> for Vec3<R> {
    fn from((x, y, z): (R, R, R)) -> Self {
        Self { x, y, z }
    }
}

impl<R: Real> From<Vec3u> for Vec3<R> {
    fn from(v: Vec3u) -> Self {
        Self::new(
            R::from_primitive(v.x()),
            R::from_primitive(v.y()),
            R::from_primitive(v.z())
        )
    }
}

impl<R: Real> AddAssign for Vec3<R> {
    fn add_assign(&mut self, other: Self) {
        self.x += other.x;
        self.y += other.y;
        self.z += other.z;
    }
}

impl<R: Real> Add for Vec3<R> {
    type Output = Self;
    fn add(self, other: Self) -> Self {
        let mut work = self.clone();
        work += other;
        work
    }
}

impl<R: Real> Sub for Vec3<R> {
    type Output = Self;
    fn sub(self, other: Self) -> Self {
        Self {
            x: self.x - other.x,
            y: self.y - other.y,
            z: self.z - other.z,
        }
    }
}

impl<R: Real> Neg for Vec3<R> {
    type Output = Self;
    fn neg(self) -> Self {
        Self {
            x: -self.x,
            y: -self.y,
            z: -self.z,
        }
    }
}

impl<R: Real> MulAssign<R> for Vec3<R> {
    fn mul_assign(&mut self, other: R) {
        self.x *= other;
        self.y *= other;
        self.z *= other;
    }
}

impl<R: Real> Mul<R> for Vec3<R> {
    type Output = Self;
    fn mul(self, other: R) -> Self {
        let mut work = self.clone();
        work *= other;
        work
    }
}

impl<R: Real> DivAssign<R> for Vec3<R> {
    fn div_assign(&mut self, other: R) {
        *self *= R::one() / other;
    }
}

impl<R: Real> Div<R> for Vec3<R> {
    type Output = Self;
    fn div(self, other: R) -> Self {
        let mut work = self.clone();
        work /= other;
        work
    }
}

impl<R: Real> Sum for Vec3<R> {
    fn sum<I: Iterator<Item = Self>>(iter: I) -> Self {
        iter.fold(Self::zero(), |a, b| a + b)
    }
}

#[cfg(feature = "fmt")]
impl<R: Real + fmt::LowerExp> fmt::LowerExp for Vec3<R> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
        fmt::Display::fmt("(", f)?;
        fmt::LowerExp::fmt(&self.x(), f)?;
        fmt::Display::fmt(", ", f)?;
        fmt::LowerExp::fmt(&self.y(), f)?;
        fmt::Display::fmt(", ", f)?;
        fmt::LowerExp::fmt(&self.z(), f)?;
        fmt::Display::fmt(")", f)
    }
}

#[cfg(feature = "fmt")]
impl<R: Real + fmt::Display> fmt::Display for Vec3<R> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
        fmt::Display::fmt("(", f)?;
        fmt::Display::fmt(&self.x(), f)?;
        fmt::Display::fmt(", ", f)?;
        fmt::Display::fmt(&self.y(), f)?;
        fmt::Display::fmt(", ", f)?;
        fmt::Display::fmt(&self.z(), f)?;
        fmt::Display::fmt(")", f)
    }
}
