use crate::real::Real;
use crate::vec::Vec3;

use core::ops::{Add, Div, Mul, Sub};

#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

#[cfg(feature = "fmt")]
use core::fmt::{self, Display};

#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, Hash, Eq, PartialEq, Ord, PartialOrd)]
pub struct Vec3u {
    x: u32,
    y: u32,
    z: u32,
}
impl Vec3u {
    pub fn new(x: u32, y: u32, z: u32) -> Self {
        Self {
            x,
            y,
            z,
        }
    }
    pub fn x(&self) -> u32 {
        self.x
    }
    pub fn y(&self) -> u32 {
        self.y
    }
    pub fn z(&self) -> u32 {
        self.z
    }
    pub fn unit_x() -> Self {
        Self::new(1, 0, 0)
    }
    pub fn unit_y() -> Self {
        Self::new(0, 1, 0)
    }
    pub fn unit_z() -> Self {
        Self::new(0, 0, 1)
    }
    /// Convert for use with ndarray indexing
    pub fn row_major_idx(&self) -> [usize; 3] {
        [self.z as _, self.y as _, self.x as _]
    }

    pub fn product_sum(&self) -> u32 {
        self.x * self.y * self.z
    }
}

impl From<(u32, u32, u32)> for Vec3u {
    fn from((x, y, z): (u32, u32, u32)) -> Self {
        Self::new(x, y, z)
    }
}

impl<R: Real> From<Vec3<R>> for Vec3u {
    fn from(v: Vec3<R>) -> Self {
        Self::new(v.x().to_f64() as _, v.y().to_f64() as _, v.z().to_f64() as _)
    }
}

impl Add<Vec3u> for Vec3u {
    type Output = Self;
    fn add(self, other: Self) -> Self {
        Self::new(self.x() + other.x(), self.y() + other.y(), self.z() + other.z())
    }
}

impl Sub<Vec3u> for Vec3u {
    type Output = Self;
    fn sub(self, other: Self) -> Self {
        Self::new(self.x() - other.x(), self.y() - other.y(), self.z() - other.z())
    }
}

impl Div<u32> for Vec3u {
    type Output = Self;
    fn div(self, other: u32) -> Self {
        Self::new(self.x()/other, self.y()/other, self.z()/other)
    }
}

impl Mul<u32> for Vec3u {
    type Output = Self;
    fn mul(self, other: u32) -> Self {
        Self::new(self.x()*other, self.y()*other, self.z()*other)
    }
}

#[cfg(feature = "fmt")]
impl Display for Vec3u {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        core::fmt::Debug::fmt(&(self.x, self.y, self.z), f)
    }
}
